from __future__ import annotations
from abc import ABC, abstractmethod
from enum import IntEnum, unique
from typing import List, Optional, Union

import numpy as np

from ..common.config import Config
from ..math.matrix import Matrix
from .geom_algo import GeomAlgo2D


class Shape(ABC):
    @unique
    class Type(IntEnum):
        BASE: int = -1
        Point: int = 0
        Polygon: int = 1
        Circle: int = 2
        Ellipse: int = 3
        Capsule: int = 4
        Edge: int = 5
        Curve: int = 6
        Sector: int = 7

    def __init__(self):
        self._type: Shape.Type = Shape.Type.BASE

    @property
    def type(self) -> Shape.Type:
        return self._type

    @type.setter
    def type(self, val: Shape.Type) -> None:
        self._type = val

    @abstractmethod
    def scale(self, factor: float) -> None:
        raise NotImplementedError

    @abstractmethod
    def contains(self, point: Matrix) -> bool:
        raise NotImplementedError

    @abstractmethod
    def center(self) -> Matrix:
        raise NotImplementedError


class ShapePrimitive():
    '''Basic Shape Description Primitive.
       Including vertices/position/angle of shape
    '''
    def __init__(self):
        self._shape: Optional[Union[Point, Polygon, Rectangle, Circle, Ellipse,
                                    Edge, Curve, Capsule, Sector]] = None
        self._xform: Matrix = Matrix([0.0, 0.0], 'vec')
        self._rot: float = 0.0

    def translate(self, src: Matrix) -> Matrix:
        assert src.shape == (2, 1)
        return Matrix.rotate_mat(self._rot) * src + self._xform


class Point(Shape):
    def __init__(self):
        super().__init__()
        self.type = self.Type.Point
        self._pos: Matrix = Matrix([0.0, 0.0], 'vec')

    @property
    def pos(self) -> Matrix:
        return self._pos

    @pos.setter
    def pos(self, pos: Matrix) -> None:
        self._pos = pos

    def scale(self, factor: float) -> None:
        self._pos *= factor

    def contains(self, point: Matrix) -> bool:
        return self._pos == point

    def center(self) -> Matrix:
        return self._pos


class Polygon(Shape):
    '''Convex polygon, not concve
    '''
    def __init__(self):
        super().__init__()
        self.type = self.Type.Polygon
        self._vertices: List[Matrix] = []

    @property
    def vertices(self) -> List[Matrix]:
        return self._vertices

    @vertices.setter
    def vertices(self, vertices: List[Matrix]) -> None:
        self._vertices = vertices
        self.update_vertices()

    def append(self, vertice: Matrix) -> None:
        raise AssertionError('dont use, otherwise make gjk algo fail! ')
        self._vertices.append(vertice)
        self.update_vertices()

    def scale(self, factor: float) -> None:
        assert len(self._vertices) > 0
        self._vertices = [v * factor for v in self._vertices]

    def contains(self, point: Matrix) -> bool:
        assert len(self._vertices) > 2

        vert_len: int = len(self._vertices)
        for i in range(vert_len - 1):
            p1: Matrix = self._vertices[i]
            p2: Matrix = self._vertices[i + 1]
            ref: Matrix = self._vertices[
                1] if i + 2 == vert_len else self._vertices[i + 2]
            # NOTE: why [1]? according to the vertices list type,
            # shape need to be closed
            if not GeomAlgo2D.is_point_on_same_side(p1, p2, ref, point):
                return False

        return True

    def center(self) -> Matrix:
        return GeomAlgo2D.calc_mass_center(self._vertices)

    def update_vertices(self) -> None:
        center_point: Matrix = self.center()
        self._vertices = [v - center_point for v in self._vertices]


class Rectangle(Polygon):
    def __init__(self, width: float = 0.0, height: float = 0.0):
        super().__init__()
        self.set_value(width, height)

    @property
    def width(self) -> float:
        return self._width

    @width.setter
    def width(self, width: float) -> None:
        self._width: float = width
        self.calc_vertices()

    @property
    def height(self) -> float:
        return self._height

    @height.setter
    def height(self, height: float) -> None:
        self._height: float = height
        self.calc_vertices()

    @property
    def vertices(self) -> List[Matrix]:
        return self._vertices

    @vertices.setter
    def vertices(self, vert: List[Matrix]) -> None:
        self._vertices = vert

    # NOTE: use _var to set val, because property 'width'
    # and 'height' call 'calc_vertices' will use _var in init
    def set_value(self, width: float, height: float) -> None:
        self._width = width
        self._height = height
        self.calc_vertices()

    def scale(self, factor: float) -> None:
        self._width *= factor
        self._height *= factor
        self.calc_vertices()

    @staticmethod
    def contain_helper(val: float, ref: float) -> bool:
        return -ref * 0.5 < val < ref * 0.5

    def contains(self, point: Matrix) -> bool:
        return self.contain_helper(point.x,
                                   self._width) and self.contain_helper(
                                       point.y, self._height)

    def calc_vertices(self) -> None:
        self._vertices: List[Matrix] = []
        self._vertices.append(
            Matrix([-self._width * 0.5, self._height * 0.5], 'vec'))
        self._vertices.append(
            Matrix([-self._width * 0.5, -self._height * 0.5], 'vec'))
        self._vertices.append(
            Matrix([self._width * 0.5, -self._height * 0.5], 'vec'))
        self._vertices.append(
            Matrix([self._width * 0.5, self._height * 0.5], 'vec'))
        self._vertices.append(
            Matrix([-self._width * 0.5, self._height * 0.5], 'vec'))


class Circle(Shape):
    def __init__(self, radius: float = 0.0):
        super().__init__()
        self.type = self.Type.Circle
        self.radius = radius

    @property
    def radius(self) -> float:
        return self._radius

    @radius.setter
    def radius(self, radius: float) -> None:
        self._radius: float = radius

    def scale(self, factor: float) -> None:
        self._radius *= factor

    def contains(self, point: Matrix) -> bool:
        return point.len_square() < self._radius * self._radius

    def center(self) -> Matrix:
        return Matrix([0.0, 0.0], 'vec')


class Ellipse(Shape):
    def __init__(self, width: float = 0.0, height: float = 0.0):
        super().__init__()
        self.type = self.Type.Ellipse
        self.set_value(width, height)

    @property
    def width(self) -> float:
        return self._width

    @width.setter
    def width(self, width: float) -> None:
        self._width: float = width

    @property
    def height(self) -> float:
        return self._height

    @height.setter
    def height(self, height: float) -> None:
        self._height: float = height

    def set_value(self, width: float, height: float) -> None:
        self._width = width
        self._height = height

    def scale(self, factor: float) -> None:
        self._width *= factor
        self._height *= factor

    def contains(self, point: Matrix) -> bool:
        return False

    def center(self) -> Matrix:
        return Matrix([0.0, 0.0], 'vec')

    def A(self) -> float:
        return self._width / 2.0

    def B(self) -> float:
        return self._height / 2.0

    def C(self) -> float:
        va = self.A()
        vb = self.B()
        return np.sqrt(va * va - vb * vb)


class Edge(Shape):
    def __init__(self):
        super().__init__()
        self.type = self.Type.Edge
        # set (1.0, 1.0) just for init calc the normal vec correctly
        # if both are (0.0, 0.0), can lead div zero error
        self.set_value(Matrix([1.0, 1.0], 'vec'), Matrix([0.0, 0.0], 'vec'))

    @property
    def start(self) -> Matrix:
        return self._start

    @start.setter
    def start(self, point: Matrix) -> None:
        self._start: Matrix = point
        self.update_normal()

    @property
    def end(self) -> Matrix:
        return self._end

    @end.setter
    def end(self, point: Matrix) -> None:
        self._end: Matrix = point
        self.update_normal()

    # the only method to init the edge value
    def set_value(self, start: Matrix, end: Matrix) -> None:
        self._start = start
        self._end = end
        self.update_normal()

    def update_normal(self) -> None:
        self._normal: Matrix = (self._end -
                                self._start).perpendicular().normal().negate()

    def scale(self, factor: float) -> None:
        self._start *= factor
        self._end *= factor

    def contains(self, point: Matrix) -> bool:
        return GeomAlgo2D.is_point_on_segment(self._start, self._end, point)

    def center(self) -> Matrix:
        return (self._start + self._end) / 2.0

    @property
    def normal(self) -> Matrix:
        return self._normal

    @normal.setter
    def normal(self, normal: Matrix) -> None:
        self._normal = normal


class Curve(Shape):
    def __init__(self):
        super().__init__()
        self.type = self.Type.Curve
        self.set_value(Matrix([0.0, 0.0], 'vec'), Matrix([0.0, 0.0], 'vec'),
                       Matrix([0.0, 0.0], 'vec'), Matrix([0.0, 0.0], 'vec'))

    @property
    def start(self) -> Matrix:
        return self._start

    @start.setter
    def start(self, point: Matrix) -> None:
        self._start: Matrix = point

    @property
    def ctrl1(self) -> Matrix:
        return self._ctrl1

    @ctrl1.setter
    def ctrl1(self, point: Matrix) -> None:
        self._ctrl1: Matrix = point

    @property
    def ctrl2(self) -> Matrix:
        return self._ctrl2

    @ctrl2.setter
    def ctrl2(self, point: Matrix) -> None:
        self._ctrl2: Matrix = point

    @property
    def end(self) -> Matrix:
        return self._end

    @end.setter
    def end(self, point: Matrix) -> None:
        self._end: Matrix = point

    def set_value(self, start: Matrix, ctrl1: Matrix, ctrl2: Matrix,
                  end: Matrix) -> None:
        self._start = start
        self._ctrl1 = ctrl1
        self._ctrl2 = ctrl2
        self._end = end

    def scale(self, factor: float) -> None:
        self._start *= factor
        self._ctrl1 *= factor
        self._ctrl2 *= factor
        self._end *= factor

    def contains(self, point: Matrix) -> bool:
        return False

    def center(self) -> Matrix:
        return Matrix([0.0, 0.0], 'vec')


class Capsule(Shape):
    def __init__(self, width: float = 0.0, height: float = 0.0):
        super().__init__()
        self.type = self.Type.Capsule
        self.set_value(width, height)

    @property
    def width(self) -> float:
        return self._width

    @width.setter
    def width(self, width: float) -> None:
        self._width: float = width

    @property
    def height(self) -> float:
        return self._height

    @height.setter
    def height(self, height: float) -> None:
        self._height: float = height

    def set_value(self, width: float, height: float) -> None:
        self._width = width
        self._height = height

    def calc_pos(self, sign: int = 1) -> List[float]:
        res: List[float] = []
        tmp: float = 0.0
        if self._width > self._height:
            tmp = self._height / 2.0
            res.append(-self._width / 2.0 + tmp)
            res.append(sign * tmp)
        else:
            tmp = self._width / 2.0
            res.append(-tmp)
            res.append(sign * (self._height / 2.0 - tmp))

        return res

    # NOTE: the capsule is made by 2 circle and 1 rectangle with
    # the radius go through the rectangle's width
    # so the top left is the circle's inscribe rectangle's top left point
    def top_left(self) -> Matrix:
        return Matrix(self.calc_pos(), 'vec')

    def bottom_left(self) -> Matrix:
        return Matrix(self.calc_pos(-1), 'vec')

    def top_right(self) -> Matrix:
        return -self.bottom_left()

    def bottom_right(self) -> Matrix:
        return -self.top_left()

    def box_vertices(self) -> List[Matrix]:
        vertices: List[Matrix] = []
        vertices.append(self.top_left())
        vertices.append(self.bottom_left())
        vertices.append(self.bottom_right())
        vertices.append(self.top_right())
        vertices.append(self.top_left())
        return vertices

    def scale(self, factor: float) -> None:
        self._width *= factor
        self._height *= factor

    @staticmethod
    def range_helper_x(p: Matrix, ref1: Matrix, ref2: Matrix,
                       idx: int) -> bool:
        if p._val[idx] - ref1._val[idx] <= Config.Epsilon and p._val[
                idx] - ref2._val[idx] >= Config.Epsilon:
            return True

        return False

    @staticmethod
    def range_helper_y(p: Matrix, idx: int, dis: float) -> bool:
        if p._val[idx] - dis <= Config.Epsilon and p._val[
                idx] + dis >= Config.Epsilon:
            return True

        return False

    @staticmethod
    def range_helper_val(p: Matrix, ref: Matrix, dis: float) -> bool:
        return (p - ref).len_square() - dis * dis <= Config.Epsilon

    def contains(self, point: Matrix) -> bool:
        anchor1: Matrix = Matrix([0.0, 0.0], 'vec')
        anchor2: Matrix = Matrix([0.0, 0.0], 'vec')
        x_len: float = 0.0
        y_len: float = 0.0

        if self._width >= self._height:
            y_len = self._height / 2.0
            x_len = self._width - self._height
            anchor1.set_value([x_len / 2.0, 0.0])
            anchor2.set_value([-x_len / 2.0, 0.0])
            if Capsule.range_helper_x(point, anchor1, anchor2,
                                      0) and Capsule.range_helper_y(
                                          point, 1, y_len):
                return True
        else:
            y_len = self._width / 2.0
            x_len = self._height - self._width
            anchor1.set_value([0.0, x_len / 2.0])
            anchor2.set_value([0.0, -x_len / 2.0])
            if Capsule.range_helper_x(point, anchor1, anchor2,
                                      1) and Capsule.range_helper_y(
                                          point, 0, y_len):
                return True

        # judge if the point in the arc part of the capsule
        if Capsule.range_helper_val(anchor1, point,
                                    y_len) or Capsule.range_helper_val(
                                        anchor2, point, y_len):
            return True

        return False

    def center(self):
        return Matrix([0.0, 0.0], 'vec')


class Sector(Shape):
    def __init__(self):
        super().__init__()
        self.type = self.Type.Sector
        self.set_value()

    def vertices(self) -> List[Matrix]:
        res: List[Matrix] = []
        res.append(Matrix([0.0, 0.0], 'vec'))
        res.append(
            Matrix.rotate_mat(self._start) * Matrix([self._radius, 0], 'vec'))
        res.append(
            Matrix.rotate_mat(self._start + self._span) *
            Matrix([self._radius, 0], 'vec'))
        res.append(Matrix([0.0, 0.0], 'vec'))
        return res

    @property
    def start(self) -> float:
        return self._start

    @start.setter
    def start(self, radian: float) -> None:
        self._start: float = radian

    @property
    def span(self) -> float:
        return self._span

    @span.setter
    def span(self, radian: float) -> None:
        self._span: float = radian

    @property
    def radius(self) -> float:
        return self._radius

    @radius.setter
    def radius(self, radius: float) -> None:
        self._radius: float = radius

    def set_value(self,
                  start: float = 0.0,
                  span: float = 0.0,
                  radius: float = 0.0) -> None:
        self._start = start
        self._span = span
        self._radius = radius

    def area(self) -> float:
        return self._span * self._radius * self._radius / 2.0

    def scale(self, factor: float) -> None:
        self._radius *= factor

    def contains(self, point: Matrix) -> bool:
        if np.isclose(point.x, 0):
            return 0 <= point.y <= self._radius
        else:
            theta: float = point.theta()
            ang_check1: bool = theta >= self._start
            ang_check2: bool = theta <= self._start + self._span
            len_check: bool = point.len_square() <= self._radius * self._radius
            return ang_check1 and ang_check2 and len_check

    def center(self) -> Matrix:
        vertices: List[Matrix] = self.vertices()
        point1: Matrix = vertices[1]
        point2: Matrix = vertices[2]
        normal: Matrix = (point1 + point2) / 2
        normal.normalize()

        point_len: float = (point1 - point2).len()
        rad_len: float = self._radius * self._span
        res: Matrix = normal * (2.0 * self._radius * point_len /
                                (3.0 * rad_len))

        return res



from typing import List


class RandomGenerator():
    # now only use the seq methods(not random) to give unique id
    # NOTE: need to make a single instance in the project
    empty_list: List[int] = []
    start_id: int = 0

    @staticmethod
    def unique() -> int:
        if len(RandomGenerator.empty_list) > 0:
            res: int = RandomGenerator.empty_list.pop()
            return res

        RandomGenerator.start_id += 1
        return RandomGenerator.start_id

    @staticmethod
    def pop(val: int):
        RandomGenerator.empty_list.append(val)



from __future__ import annotations
from typing import List, Dict, Optional, Union, Tuple

import numpy as np

from ...math.matrix import Matrix
from ...common.config import Config
from ...dynamics.body import Body
from .aabb import AABB


class DBVT():
    '''Dynamic Bounding Volume Tree
    This is implemented by dynamic array-arranged.
    '''
    class Node():
        def __init__(self):
            self._body: Optional[Body] = None
            self._aabb: AABB = AABB()
            self._parent_idx: int = -1
            self._left_idx: int = -1
            self._right_idx: int = -1

        def is_leaf(self) -> bool:
            return self._left_idx == -1 and self._right_idx == -1

        def is_branch(self) -> bool:
            child_ck: bool = self._left_idx != -1 and self._right_idx != -1
            return self._parent_idx != -1 and child_ck

        def is_root(self) -> bool:
            child_ck: bool = self._left_idx != -1 and self._right_idx != -1
            return self._parent_idx == -1 and child_ck

        def is_empty(self) -> bool:
            return self._aabb.is_empty()

        def clear(self) -> None:
            self._body = None
            self._aabb.clear()
            self._parent_idx = -1
            self._left_idx = -1
            self._right_idx = -1

    def __init__(self):
        self._fat_expansion_factor: float = 0.5
        self._root_idx: int = -1
        self._tree: List[DBVT.Node] = []
        self._empty_list: List[int] = []
        self._body_table: Dict[Body, int] = {}

    def query(self, val: Union[Body, AABB]) -> List[Body]:
        res: List[Body] = []
        aabb: Optional[AABB] = None

        if isinstance(val, Body):
            aabb = AABB.from_body(val)
        elif isinstance(val, AABB):
            aabb = val

        assert aabb is not None
        self._query_nodes(self._root_idx, aabb, res)
        return res

    def raycast(self, start: Matrix, dirn: Matrix) -> List[Body]:
        res: List[Body] = []
        self._raycast(res, self._root_idx, start, dirn)
        return res

    def generate(self) -> List[Tuple[Body, Body]]:
        pairs: List[Tuple[Body, Body]] = []
        self._generate(self._root_idx, pairs)
        return pairs

    def insert(self, body: Body) -> None:
        new_node_idx: int = self._allocate_node()
        self._tree[new_node_idx]._body = body
        self._tree[new_node_idx]._aabb = AABB.from_body(body)
        # expand outline a litte
        self._tree[new_node_idx]._aabb.expand(self._fat_expansion_factor)
        self._body_table[body] = new_node_idx

        if self._root_idx == -1:
            self._root_idx = new_node_idx
            return

        if self._tree[self._root_idx].is_leaf():
            self._root_idx = self._merge(new_node_idx, self._root_idx)
            return

        # calc cost
        target_idx: int = self._calc_lowest_cost_node(new_node_idx)
        if target_idx == self._root_idx:
            self._root_idx = self._merge(new_node_idx, target_idx)
            self._balance(self._root_idx)
            return

        target_parent_idx: int = self._tree[target_idx]._parent_idx
        self._separate(target_idx, target_parent_idx)
        box_idx: int = self._merge(new_node_idx, target_idx)
        self._join(box_idx, target_parent_idx)
        self._upgrade(box_idx)
        self._balance(self._root_idx)

    def remove(self, body: Body) -> None:
        is_find: bool = False
        for v in self._body_table:
            if v == body:
                is_find = True
                break

        if not is_find:
            return

        parent_idx: int = self._tree[self._body_table[body]]._parent_idx
        if parent_idx == -1 and self._tree[self._body_table[body]].is_leaf():
            self._root_idx = -1
            self._remove(self._body_table[body])
            del self._body_table[body]
            return

        another_child: int = self._tree[parent_idx]._right_idx if self._tree[
            parent_idx]._left_idx == self._body_table[body] else self._tree[
                parent_idx]._left_idx
        self._remove(self._body_table[body])
        self._elevate(another_child)
        self._upgrade(another_child)
        del self._body_table[body]

    def clear_all(self) -> None:
        self._tree = []
        self._empty_list = []
        self._body_table = {}
        self._root_idx = -1

    def update(self, body: Body) -> None:
        is_find: bool = False
        for v in self._body_table:
            if v == body:
                is_find = True
                break

        if not is_find:
            return

        thin: AABB = AABB.from_body(body)
        thin.expand(0.1)
        if not thin.is_subset(self._tree[self._body_table[body]]._aabb):
            self._extract(self._body_table[body])
            self.insert(body)

    def tree(self) -> List[Node]:
        return self._tree

    def root_index(self) -> int:
        return self._root_idx

    def _query_nodes(self, node_idx: int, aabb: AABB, res: List[Body]):
        if node_idx == -1:
            return

        overlap: bool = self._tree[node_idx]._aabb.collide(aabb) or self._tree[
            node_idx]._aabb.is_subset(aabb) or aabb.is_subset(
                self._tree[node_idx]._aabb)

        if not overlap:
            return

        if self._tree[node_idx].is_leaf():
            tmp: Optional[Body] = self._tree[node_idx]._body
            assert tmp is not None
            res.append(tmp)
            return

        if self._tree[node_idx].is_branch() or self._tree[node_idx].is_root():
            self._query_nodes(self._tree[node_idx]._left_idx, aabb, res)
            self._query_nodes(self._tree[node_idx]._right_idx, aabb, res)

    # NOTE: now use the List to return the var
    def _traverse_lowest_cost(self, node_idx: int, box_idx: int,
                              cost: List[float], final_idx: List[int]) -> None:
        def _accumulate_cost(node_idx: int, box_idx: int) -> float:
            if self._tree[node_idx].is_leaf():
                return inherit_cost + AABB.unite(
                    self._tree[node_idx]._aabb,
                    self._tree[box_idx]._aabb).surface_area()

            return self._delta_cost(node_idx, box_idx) + inherit_cost

        if self._tree[box_idx].is_leaf() and not self._tree[box_idx].is_root():
            final_idx[0] = box_idx
            return

        area: float = self._tree[box_idx]._aabb.surface_area()
        union_area: float = AABB.unite(
            self._tree[node_idx]._aabb,
            self._tree[box_idx]._aabb).surface_area()

        cost[0] = 2.0 * area
        inherit_cost: float = 2.0 * (union_area - area)

        left_idx: int = self._tree[box_idx]._left_idx
        right_idx: int = self._tree[box_idx]._right_idx
        left_cost: float = _accumulate_cost(node_idx, left_idx)
        right_cost: float = _accumulate_cost(node_idx, right_idx)
        lowest_cost: float = 0.0
        lowest_const_idx: int = 0

        if cost[0] < left_cost and cost[0] < right_cost:
            final_idx[0] = box_idx
            return

        if left_cost > right_cost:
            lowest_cost = right_cost
            lowest_const_idx = right_idx
        else:
            lowest_cost = left_cost
            lowest_const_idx = left_idx

        final_idx[0] = lowest_const_idx
        self._traverse_lowest_cost(node_idx, lowest_const_idx, [lowest_cost],
                                   final_idx)

    def _raycast(self, res: List[Body], node_idx: int, p: Matrix,
                 d: Matrix) -> None:
        if node_idx < 0:
            return

        if self._tree[node_idx]._aabb.raycast(p, d):
            if self._tree[node_idx].is_leaf():
                tmp: Optional[Body] = self._tree[node_idx]._body
                assert tmp is not None
                res.append(tmp)
            else:
                self._raycast(res, self._tree[node_idx]._left_idx, p, d)
                self._raycast(res, self._tree[node_idx]._right_idx, p, d)

    def _generate(self, node_idx: int, pairs: List[Tuple[Body, Body]]) -> None:
        if node_idx < 0 or self._tree[node_idx].is_leaf():
            return

        left_idx: int = self._tree[node_idx]._left_idx
        right_idx: int = self._tree[node_idx]._right_idx
        res: bool = self._tree[left_idx]._aabb.collide(
            self._tree[right_idx]._aabb)

        if res:
            self._generate2(left_idx, right_idx, pairs)

        self._generate(left_idx, pairs)
        self._generate(right_idx, pairs)

    def _generate2(self, left_idx: int, right_idx: int,
                   pairs: List[Tuple[Body, Body]]) -> None:
        if left_idx < 0 or right_idx < 0:
            return

        left_aabb: AABB = self._tree[left_idx]._aabb
        right_aabb: AABB = self._tree[right_idx]._aabb
        res: bool = left_aabb.collide(right_aabb) or left_aabb.is_subset(
            right_aabb) or right_aabb.is_subset(left_aabb)

        if not res:
            return

        if self._tree[left_idx].is_leaf() and self._tree[right_idx].is_leaf():
            tmpl: Optional[Body] = self._tree[left_idx]._body
            tmpr: Optional[Body] = self._tree[right_idx]._body
            assert tmpl is not None
            assert tmpr is not None

            if tmpl.bitmask & tmpr.bitmask:
                # if AABB of A & B overlap
                if AABB.from_body(tmpl).collide(AABB.from_body(tmpr)):
                    pairs.append((tmpl, tmpr))

        if self._tree[left_idx].is_leaf() and self._tree[right_idx].is_branch(
        ):
            self._generate2(left_idx, self._tree[right_idx]._left_idx, pairs)
            self._generate2(left_idx, self._tree[right_idx]._right_idx, pairs)

        if self._tree[right_idx].is_leaf() and self._tree[left_idx].is_branch(
        ):
            self._generate2(right_idx, self._tree[left_idx]._left_idx, pairs)
            self._generate2(right_idx, self._tree[left_idx]._right_idx, pairs)

        if self._tree[left_idx].is_branch(
        ) and self._tree[right_idx].is_branch():
            self._generate2(self._tree[left_idx]._left_idx, right_idx, pairs)
            self._generate2(self._tree[left_idx]._right_idx, right_idx, pairs)

    def _extract(self, target_idx: int) -> None:
        another_child_index: int = -1
        tmp: Optional[Body] = None

        if target_idx == self._root_idx:
            self._root_idx = -1

            tmp = self._tree[target_idx]._body
            assert tmp is not None
            self._body_table[tmp] = -1
            self._remove(target_idx)
            return

        if self._tree[target_idx]._parent_idx == self._root_idx:
            another_child_index = self._tree[
                self._root_idx]._right_idx if self._tree[
                    self._root_idx]._left_idx == target_idx else self._tree[
                        self._root_idx]._left_idx

            self._separate(target_idx, self._root_idx)
            self._elevate(another_child_index)

            tmp = self._tree[target_idx]._body
            assert tmp is not None
            self._body_table[tmp] = -1
            self._remove(target_idx)
            return

        parent_idx: int = self._tree[target_idx]._parent_idx
        another_child_index = self._tree[parent_idx]._right_idx if self._tree[
            parent_idx]._left_idx == target_idx else self._tree[
                parent_idx]._left_idx

        self._separate(target_idx, parent_idx)
        self._elevate(another_child_index)

        tmp = self._tree[target_idx]._body
        assert tmp is not None
        self._body_table[tmp] = -1
        self._remove(target_idx)

    def _merge(self, node_idx: int, leaf_idx: int) -> int:
        parent_idx: int = self._allocate_node()
        self._tree[leaf_idx]._parent_idx = parent_idx
        self._tree[node_idx]._parent_idx = parent_idx
        self._tree[parent_idx]._left_idx = leaf_idx
        self._tree[parent_idx]._right_idx = node_idx
        self._tree[parent_idx]._aabb = AABB.unite(self._tree[node_idx]._aabb,
                                                  self._tree[leaf_idx]._aabb)
        return parent_idx

    def _ll(self, node_idx: int) -> None:
        if node_idx == -1 or self._tree[node_idx].is_root():
            return

        parent_idx: int = -1
        grand_idx: int = -1
        right_idx: int = -1
        if self._tree[node_idx]._parent_idx == self._root_idx:
            parent_idx = self._tree[node_idx]._parent_idx
            right_idx = self._tree[node_idx]._right_idx
            self._separate(node_idx, parent_idx)
            self._separate(right_idx, node_idx)
            self._join(right_idx, parent_idx)
            self._join(parent_idx, node_idx)
            self._root_idx = node_idx
            self._upgrade(parent_idx)
            return

        parent_idx = self._tree[node_idx]._parent_idx
        grand_idx = self._tree[parent_idx]._parent_idx
        right_idx = self._tree[node_idx]._right_idx

        self._separate(parent_idx, grand_idx)
        self._separate(node_idx, parent_idx)
        self._separate(right_idx, node_idx)

        self._join(right_idx, parent_idx)
        self._join(parent_idx, node_idx)
        self._join(node_idx, grand_idx)
        self._upgrade(parent_idx)

    def _rr(self, node_idx: int) -> None:
        if node_idx == -1 or self._tree[node_idx].is_root():
            return

        parent_idx: int = -1
        grand_idx: int = -1
        left_idx: int = -1
        if self._tree[node_idx]._parent_idx == self._root_idx:
            parent_idx = self._tree[node_idx]._parent_idx
            left_idx = self._tree[node_idx]._left_idx
            self._separate(node_idx, parent_idx)
            self._separate(left_idx, node_idx)
            self._join(left_idx, parent_idx)
            self._join(parent_idx, node_idx)
            self._root_idx = node_idx
            self._upgrade(parent_idx)
            return

        parent_idx = self._tree[node_idx]._parent_idx
        grand_idx = self._tree[parent_idx]._parent_idx
        left_idx = self._tree[node_idx]._left_idx

        self._separate(parent_idx, grand_idx)
        self._separate(node_idx, parent_idx)
        self._separate(left_idx, node_idx)
        self._join(left_idx, parent_idx)
        self._join(parent_idx, node_idx)
        self._join(node_idx, grand_idx)
        self._upgrade(parent_idx)

    def _balance(self, target_idx: int) -> None:
        if target_idx == -1 or self._tree[target_idx].is_leaf():
            return

        left_height: int = self._height(self._tree[target_idx]._left_idx)
        right_height: int = self._height(self._tree[target_idx]._right_idx)
        if np.fabs(left_height - right_height) <= 1:
            return

        # left unbalance
        if left_height > right_height:
            ll_height: int = self._height(
                self._tree[self._tree[target_idx]._left_idx]._left_idx)
            lr_height: int = self._height(
                self._tree[self._tree[target_idx]._left_idx]._right_idx)

            # LR case
            if ll_height < lr_height:
                self._rr(
                    self._tree[self._tree[target_idx]._left_idx]._right_idx)
            else:
                self._ll(
                    self._tree[self._tree[target_idx]._left_idx]._left_idx)

            self._ll(self._tree[target_idx]._left_idx)

        else:  # right unbalance
            rr_height: int = self._height(
                self._tree[self._tree[target_idx]._right_idx]._right_idx)
            rl_height: int = self._height(
                self._tree[self._tree[target_idx]._right_idx]._left_idx)

            # RL case
            if rr_height < rl_height:
                self._ll(
                    self._tree[self._tree[target_idx]._right_idx]._left_idx)
            else:
                self._rr(
                    self._tree[self._tree[target_idx]._right_idx]._right_idx)

            self._rr(self._tree[target_idx]._right_idx)

        self._balance(self._tree[target_idx]._left_idx)
        self._balance(self._tree[target_idx]._right_idx)
        self._balance(self._tree[target_idx]._parent_idx)

    def _separate(self, source_idx: int, parent_idx: int) -> None:
        if source_idx < 0 or parent_idx < 0:
            return

        if self._tree[parent_idx]._left_idx == source_idx:
            self._tree[parent_idx]._left_idx = -1
        elif self._tree[parent_idx]._right_idx == source_idx:
            self._tree[parent_idx]._right_idx = -1
        self._tree[source_idx]._parent_idx = -1

    def _join(self, node_idx: int, box_idx: int) -> None:
        if node_idx < 0 or box_idx < 0:
            return

        if self._tree[box_idx]._left_idx == -1:
            self._tree[box_idx]._left_idx = node_idx
        elif self._tree[box_idx]._right_idx == -1:
            self._tree[box_idx]._right_idx = node_idx
        self._tree[node_idx]._parent_idx = box_idx

    def _remove(self, target_idx: int) -> None:
        self._tree[target_idx].clear()
        self._empty_list.append(target_idx)

    def _elevate(self, target_idx: int) -> None:
        if self._tree[target_idx]._parent_idx == self._root_idx:
            self._remove(self._root_idx)
            self._root_idx = target_idx
            self._tree[target_idx]._parent_idx = -1
            return

        parent_idx: int = self._tree[target_idx]._parent_idx
        grand_idx: int = self._tree[parent_idx]._parent_idx
        self._separate(target_idx, parent_idx)
        self._separate(parent_idx, grand_idx)
        self._join(target_idx, grand_idx)
        self._remove(parent_idx)

    def _upgrade(self, node_idx: int) -> None:
        if node_idx < 0 or self._tree[node_idx].is_leaf():
            return

        self._tree[node_idx]._aabb = AABB.unite(
            self._tree[self._tree[node_idx]._left_idx]._aabb,
            self._tree[self._tree[node_idx]._right_idx]._aabb)

        self._upgrade(self._tree[node_idx]._parent_idx)

    def _calc_lowest_cost_node(self, node_idx: int) -> int:
        lowest_cost: List[float] = [Config.Max]
        target_idx: List[int] = [-1]

        # start traverse lowest cost node
        # the lowest_cost is not returned
        self._traverse_lowest_cost(node_idx, self._root_idx, lowest_cost,
                                   target_idx)
        return target_idx[0]

    def _total_cost(self, node_idx: int, leaf_idx: int) -> float:
        total_cost: float = AABB.unite(
            self._tree[node_idx]._aabb,
            self._tree[leaf_idx]._aabb).surface_area()
        cur_idx: int = self._tree[leaf_idx]._parent_idx

        while cur_idx != -1:
            total_cost += self._delta_cost(node_idx, cur_idx)
            cur_idx = self._tree[cur_idx]._parent_idx

        return total_cost

    def _delta_cost(self, node_idx: int, box_idx: int) -> float:
        return AABB.unite(self._tree[box_idx]._aabb,
                          self._tree[node_idx]._aabb).surface_area(
                          ) - self._tree[box_idx]._aabb.surface_area()

    def _allocate_node(self) -> int:
        if len(self._empty_list) > 0:
            return self._empty_list.pop()

        self._tree.append(DBVT.Node())
        return len(self._tree) - 1

    def _height(self, target_idx: int) -> int:
        return 0 if target_idx < 0 else np.fmax(
            self._height(self._tree[target_idx]._left_idx),
            self._height(self._tree[target_idx]._right_idx)) + 1



from typing import List, Optional, Tuple, cast

import numpy as np

from ...math.matrix import Matrix
from ...geometry.shape import Capsule, Edge, Polygon, Sector, Shape
from ...geometry.shape import ShapePrimitive
from ...geometry.geom_algo import GeomAlgo2D
from .gjk import GJK, PointPair


class ContactGenerator():
    class ClipEdge():
        def __init__(self):
            self._p1: Matrix = Matrix([0.0, 0.0], 'vec')
            self._p2: Matrix = Matrix([0.0, 0.0], 'vec')
            self._normal: Matrix = Matrix([0.0, 0.0], 'vec')

        def is_empty(self) -> bool:
            return self._p1.is_origin() and self._p2.is_origin()

    @staticmethod
    def dump_vertices(prim: ShapePrimitive) -> List[Matrix]:
        vertices: List[Matrix] = []

        assert prim._shape is not None
        if prim._shape.type == Shape.Type.Capsule:
            cap: Capsule = cast(Capsule, prim._shape)
            vertices = cap.box_vertices()

        elif prim._shape.type == Shape.Type.Polygon:
            poly: Polygon = cast(Polygon, prim._shape)
            vertices = poly.vertices

        elif prim._shape.type == Shape.Type.Edge:
            edg: Edge = cast(Edge, prim._shape)
            vertices.append(edg.start)
            vertices.append(edg.end)

        elif prim._shape.type == Shape.Type.Sector:
            sec: Sector = cast(Sector, prim._shape)
            vertices = sec.vertices()

        vertices = [prim.translate(v) for v in vertices]
        return vertices

    @staticmethod
    def find_clip_edge(vertices: List[Matrix], idx: int,
                       normal: Matrix) -> ClipEdge:
        edg1 = ContactGenerator.ClipEdge()
        edg2 = ContactGenerator.ClipEdge()
        edg1._p2 = vertices[idx]
        edg2._p1 = vertices[idx]

        if idx == 0:
            edg1._p1 = vertices[len(vertices) - 2]
            edg2._p2 = vertices[idx + 1]

        elif idx == len(vertices) - 1:
            edg1._p1 = vertices[idx - 1]
            edg2._p2 = vertices[1]

        else:
            edg1._p1 = vertices[idx - 1]
            edg2._p2 = vertices[idx + 1]

        # compare which is closest to normal
        final_edg: ContactGenerator.ClipEdge = ContactGenerator.ClipEdge()
        p: Matrix = Matrix([0.0, 0.0], 'vec')

        tmp_val1: float = np.fabs((edg1._p2 - edg1._p1).dot(normal))
        tmp_val2: float = np.fabs((edg2._p2 - edg2._p1).dot(normal))
        if tmp_val1 >= tmp_val2:
            final_edg = edg2
            p = (edg2._p2 - edg2._p1).normal().perpendicular()
            if GeomAlgo2D.is_point_on_same_side(edg2._p1, edg2._p2, edg1._p1,
                                                edg2._p1 + p):
                final_edg._normal = p
            else:
                final_edg._normal = -p
        else:
            final_edg = edg1
            p = (edg1._p2 - edg1._p1).normal().perpendicular()
            if GeomAlgo2D.is_point_on_same_side(edg1._p1, edg1._p2, edg2._p2,
                                                edg1._p1 + p):
                final_edg._normal = p
            else:
                final_edg._normal = -p

        return final_edg

    @staticmethod
    def dump_clip_edge(prim: ShapePrimitive, vertices: List[Matrix],
                       normal: Matrix) -> ClipEdge:
        edg: ContactGenerator.ClipEdge = ContactGenerator.ClipEdge()

        if len(vertices) == 2:
            edg._p1 = vertices[0]
            edg._p2 = vertices[1]

            assert prim._shape is not None
            if prim._shape.type == Shape.Type.Edge:
                edg._normal = cast(Edge, prim._shape).normal

        else:
            support, idx = GJK.find_farthest_point2(vertices, normal)
            edg = ContactGenerator.find_clip_edge(vertices, idx, normal)

        return edg

    @staticmethod
    def recognize(prima: ShapePrimitive, primb: ShapePrimitive,
                  normal: Matrix) -> Tuple[ClipEdge, ClipEdge]:
        assert prima._shape is not None
        assert primb._shape is not None
        typea = prima._shape.type
        typeb = primb._shape.type

        tmp_shape: List[Shape.Type] = [
            Shape.Type.Point, Shape.Type.Circle, Shape.Type.Ellipse
        ]

        for s in tmp_shape:
            if typea == s or typeb == s:
                return (ContactGenerator.ClipEdge(),
                        ContactGenerator.ClipEdge())

        verta: List[Matrix] = ContactGenerator.dump_vertices(prima)
        vertb: List[Matrix] = ContactGenerator.dump_vertices(primb)
        edga = ContactGenerator.dump_clip_edge(prima, verta, -normal)
        edgb = ContactGenerator.dump_clip_edge(primb, vertb, normal)

        return (edga, edgb)

    @staticmethod
    def clip(clip_edga: ClipEdge, clip_edgb: ClipEdge,
             normal: Matrix) -> List[PointPair]:
        res: List[PointPair] = []

        if clip_edga.is_empty() or clip_edgb.is_empty():
            return res

        # find reference edge
        tmp1: Matrix = Matrix([0.0, 0.0], 'vec')
        tmp2: Matrix = Matrix([0.0, 0.0], 'vec')
        tmp1.set_value(clip_edga._p1 - clip_edga._p2)
        tmp2.set_value(clip_edgb._p1 - clip_edgb._p2)
        d1: float = tmp1.dot(normal)
        d2: float = tmp2.dot(normal)

        ref_edg: ContactGenerator.ClipEdge = clip_edga
        incident_edge: ContactGenerator.ClipEdge = clip_edgb
        swap: bool = False

        if np.fabs(d1) > np.fabs(d2):
            # edge B is reference edge
            ref_edg = clip_edgb
            incident_edge = clip_edga
            swap = True

        # 1. clip left region
        u: Matrix = (ref_edg._p2 - ref_edg._p1).normal()
        ref_anchor1: Matrix = u.perpendicular() + ref_edg._p1

        if not GeomAlgo2D.is_point_on_same_side(
                ref_edg._p1, ref_anchor1, ref_edg._p2, incident_edge._p1):

            incident_edge._p1 = GeomAlgo2D.line_intersection(
                ref_edg._p1, ref_anchor1, incident_edge._p1, incident_edge._p2)

        if not GeomAlgo2D.is_point_on_same_side(
                ref_edg._p1, ref_anchor1, ref_edg._p2, incident_edge._p2):

            incident_edge._p2 = GeomAlgo2D.line_intersection(
                ref_edg._p1, ref_anchor1, incident_edge._p1, incident_edge._p2)

        # 2. clip right region
        u.negate()
        ref_anchor2: Matrix = u.perpendicular() + ref_edg._p2
        if not GeomAlgo2D.is_point_on_same_side(
                ref_edg._p2, ref_anchor2, ref_edg._p1, incident_edge._p1):

            incident_edge._p1 = GeomAlgo2D.line_intersection(
                ref_edg._p2, ref_anchor2, incident_edge._p1, incident_edge._p2)

        if not GeomAlgo2D.is_point_on_same_side(
                ref_edg._p2, ref_anchor2, ref_edg._p1, incident_edge._p2):

            incident_edge._p2 = GeomAlgo2D.line_intersection(
                ref_edg._p2, ref_anchor2, incident_edge._p1, incident_edge._p2)

        # 3. clip normal region
        ref_anchor3: Matrix = (ref_edg._p2 +
                               ref_edg._p1) / 2.0 + ref_edg._normal

        p1_on_clip_area: bool = GeomAlgo2D.is_point_on_same_side(
            ref_edg._p1, ref_edg._p2, ref_anchor3, incident_edge._p1)

        p2_on_clip_area: bool = GeomAlgo2D.is_point_on_same_side(
            ref_edg._p1, ref_edg._p2, ref_anchor3, incident_edge._p2)

        if not (p1_on_clip_area and p2_on_clip_area):
            return res

        # p1 inside, p2 ouside
        if p1_on_clip_area and not p2_on_clip_area:
            incident_edge._p2 = GeomAlgo2D.line_intersection(
                ref_edg._p1, ref_edg._p2, incident_edge._p1, incident_edge._p2)

        # p1 outisde, p2 inside
        if not p1_on_clip_area and p2_on_clip_area:
            incident_edge._p1 = GeomAlgo2D.line_intersection(
                ref_edg._p1, ref_edg._p2, incident_edge._p1, incident_edge._p2)

        # p1 and p2 are inside, clip nothing, just go to project
        # 4. project to reference edge
        pp1: Matrix = GeomAlgo2D.point_to_line_segment(ref_edg._p1,
                                                       ref_edg._p2,
                                                       incident_edge._p1)
        pp2: Matrix = GeomAlgo2D.point_to_line_segment(ref_edg._p1,
                                                       ref_edg._p2,
                                                       incident_edge._p2)

        pair1: PointPair = PointPair()
        pair2: PointPair = PointPair()

        if not swap:
            pair1._pa = pp1
            pair1._pb = incident_edge._p1
            pair2._pa = pp2
            pair2._pb = incident_edge._p2
        else:
            pair1._pa = incident_edge._p1
            pair1._pb = pp1
            pair2._pa = incident_edge._p2
            pair2._pb = pp2

        res.append(pair1)
        res.append(pair2)
        return res



from typing import List, Dict, Optional, Tuple

import numpy as np

from .gjk import Simplex, Minkowski, GJK
from ...math.matrix import Matrix
from ...geometry.shape import ShapePrimitive
from ...geometry.geom_algo import GeomAlgo2D


class MPR():
    '''Minkowski Portal Refinement
    
    The implementation of MPR still remain some bugs when computing deeper object penetration.
    It may causes error penetration depth and normal.
    '''
    @staticmethod
    def discover(prima: ShapePrimitive,
                 primb: ShapePrimitive) -> Tuple[Matrix, Simplex]:
        '''Discover a portal for next collision test

        Parameters
        ----------
        prima : ShapePrimitive
            primitive a
        primb : ShapePrimitive
            primitive b

        Returns
        -------
        Tuple[Matrix, Simplex]
            return a vector from center of simplex to origin, an initial simplex
        '''

        simplex: Simplex = Simplex()
        ctra: Matrix = Matrix.rotate_mat(prima._rot) * prima._shape.center()
        ctrb: Matrix = Matrix.rotate_mat(primb._rot) * primb._shape.center()
        origin: Matrix = primb._xform - prima._xform
        v0: Minkowski = Minkowski(ctra + prima._xform, ctrb + primb._xform)
        dirn: Matrix = ctrb - ctra + origin

        if dirn == Matrix([0.0, 0.0], 'vec'):
            dirn.set_value([1.0, 1.0])

        v1: Minkowski = GJK.support(prima, primb, dirn)
        dirn = GJK.calc_direction_by_edge(v0._res, v1._res, True)
        v2: Minkowski = GJK.support(prima, primb, dirn)

        simplex._vertices.append(v0)
        simplex._vertices.append(v1)
        simplex._vertices.append(v2)

        return (ctrb - ctra + origin, simplex)

    @staticmethod
    def refine(prima: ShapePrimitive,
               primb: ShapePrimitive,
               src: Simplex,
               center_to_origin: Matrix,
               iter_val: int = 50) -> Tuple[bool, Simplex]:
        '''Refine portal close to origin

        Parameters
        ----------
        prima : ShapePrimitive
            primitive a
        primb : ShapePrimitive
            primitive b
        src : Simplex
            init simplex
        center_to_origin : Matrix
            none
        iter_val : int, optional
            iter num, by default 50

        Returns
        -------
        Tuple[boo, Simplex]
            return if there is a collision, the final simplex
        '''

        simplex: Simplex = src
        is_colliding: bool = False
        v1: Matrix = Matrix([0.0, 0.0], 'vec')
        v2: Matrix = Matrix([0.0, 0.0], 'vec')
        dirn: Matrix = Matrix([0.0, 0.0], 'vec')

        for i in range(iter_val):
            v1 = simplex._vertices[1]._res
            v2 = simplex._vertices[2]._res
            dirn = GJK.calc_direction_by_edge(v1, v2, True)

            if (dirn.dot(center_to_origin) < 0):
                dirn.negate()
                is_colliding = True

            new_vertex: Minkowski = GJK.support(prima, primb, dirn)

            if v1 == new_vertex._res or v2 == new_vertex._res:
                break

            dist13: float = GeomAlgo2D.point_to_line_segment(
                v1, new_vertex._res, [0.0, 0.0]).len_square()
            dist23: float = GeomAlgo2D.point_to_line_segment(
                v2, new_vertex._res, [0.0, 0.0]).len_square()

            if dist13 < dist23:
                simplex._vertices[2] = new_vertex
            else:
                simplex._vertices[1] = new_vertex

        return (is_colliding, simplex)



from typing import List, Optional

import numpy as np

from ...math.matrix import Matrix
from ..body import Body
from .joint import Joint, JointType


class DistanceJointPrimitive():
    def __init__(self):
        self._bodya: Optional[Body] = None
        self._local_pointa: Matrix = Matrix([0.0, 0.0], 'vec')
        self._target_point: Matrix = Matrix([0.0, 0.0], 'vec')
        self._normal: Matrix = Matrix([0.0, 0.0], 'vec')
        self._bias_factor: float = 0.3
        self._bias: float = 0.0
        self._dist_min: float = 0.0
        self._dist_max: float = 0.0
        self._eff_mass: float = 0.0
        self._accum_impulse: float = 0.0


class DistanceConstraintPrimitive():
    def __init__(self):
        self._bodya: Optional[Body] = None
        self._bodyb: Optional[Body] = None
        self._nearest_pa: Matrix = Matrix([0.0, 0.0], 'vec')
        self._nearest_pb: Matrix = Matrix([0.0, 0.0], 'vec')
        self._ra: Matrix = Matrix([0.0, 0.0], 'vec')
        self._rb: Matrix = Matrix([0.0, 0.0], 'vec')
        self._bias: Matrix = Matrix([0.0, 0.0], 'vec')
        self._eff_mass: Matrix = Matrix([0.0, 0.0, 0.0, 0.0])
        self._impulse: Matrix = Matrix([0.0, 0.0], 'vec')
        self._force_max: float = 200.0


class DistanceJoint(Joint):
    def __init__(self,
                 prim: DistanceJointPrimitive = DistanceJointPrimitive()):
        super().__init__()
        self._type: JointType = JointType.Distance
        self._prim: DistanceJointPrimitive = prim
        self._factor: float = 0.4

    def set_value(self, prim: DistanceJointPrimitive) -> None:
        self._prim = prim

    def prepare(self, dt: float) -> None:
        assert self._prim._dist_min <= self._prim._dist_max
        assert self._prim._bodya is not None

        bodya: Body = self._prim._bodya
        pa: Matrix = bodya.to_world_point(self._prim._local_pointa)
        ra: Matrix = pa - bodya.pos
        pb: Matrix = self._prim._target_point
        im_a: float = self._prim._bodya.inv_mass
        ii_a: float = self._prim._bodya.inv_inertia
        error: Matrix = pb - pa
        val_len: float = error.len()
        c: float = 0.0

        self._prim._normal = error.normal()
        if val_len < self._prim._dist_min:
            c = self._prim._dist_min - val_len
            self._prim._normal.negate()

        elif val_len > self._prim._dist_max:
            c = val_len - self._prim._dist_max

        else:
            self._prim._accum_impulse = 0.0
            self._prim._normal.clear()
            self._prim._bias = 0.0
            return

        if self._prim._bodya.vel.dot(self._prim._normal) > 0:
            self._prim._accum_impulse = 0.0
            self._prim._normal.clear()
            self._prim._bias = 0.0
            return

        rn_a: float = self._prim._normal.dot(ra)
        self._prim._eff_mass = 1.0 / (im_a + ii_a * rn_a * rn_a)
        self._prim._bias = self._prim._bias_factor * c / dt

    def solve_velocity(self, dt: float) -> None:
        if np.isclose(self._prim._bias, 0):
            return

        assert self._prim._bodya is not None

        ra: Matrix = self._prim._bodya.to_world_point(
            self._prim._local_pointa) - self._prim._bodya.pos
        va: Matrix = self._prim._bodya.vel + Matrix.cross_product2(
            self._prim._bodya.ang_vel, ra)
        dv: Matrix = va
        jv: float = self._prim._normal.dot(dv)
        jvb: float = -jv + self._prim._bias
        lambda_n: float = self._prim._eff_mass * jvb

        old_impulse: float = self._prim._accum_impulse
        self._prim._accum_impulse = np.fmax(old_impulse + lambda_n, 0)
        lambda_n = self._prim._accum_impulse - old_impulse

        impulse: Matrix = self._prim._normal * lambda_n
        self._prim._bodya.apply_impulse(impulse, ra)

    def solve_position(self, dt: float) -> None:
        pass

    @property
    def prim(self) -> DistanceJointPrimitive:
        return self._prim


class DistanceConstraint(Joint):
    def __init__(
        self,
        prim: DistanceConstraintPrimitive = DistanceConstraintPrimitive()):
        super().__init__()
        self._prim: DistanceConstraintPrimitive = prim
        self._factor: float = 0.1

    def prepare(self, dt: float) -> None:
        if self._prim._bodya is None or self._prim._bodyb is None:
            return

        bodya: Body = self._prim._bodya
        bodyb: Body = self._prim._bodyb

        im_a: float = bodya.inv_mass
        ii_a: float = bodya.inv_inertia
        im_b: float = bodyb.inv_mass
        ii_b: float = bodyb.inv_inertia

        self._prim._ra = self._prim._nearest_pa - bodya.pos
        self._prim._rb = self._prim._nearest_pb - bodyb.pos
        ra: Matrix = self._prim._ra
        rb: Matrix = self._prim._rb
        error: Matrix = self._prim._nearest_pa - self._prim._nearest_pb

        k: Matrix = Matrix([0.0, 0.0, 0.0, 0.0])
        data_arr: List[float] = []
        data_arr.append(im_a + ra.y * ra.y * ii_a + im_b + rb.y * rb.y * ii_b)
        data_arr.append(-ra.x * ra.y * ii_a - rb.x * rb.y * ii_b)
        data_arr.append(data_arr[1])
        data_arr.append(im_a + ra.x * ra.x * ii_a + im_b + rb.x * rb.x * ii_b)
        k.set_value(data_arr)

        self._prim._bias = error * self._factor
        self._prim._eff_mass = k.invert()

    def solve_velocity(self, dt: float) -> None:
        if self._prim._bodya is None or self._prim._bodyb is None:
            return

        va: Matrix = self._prim._bodya.vel + Matrix.cross_product2(
            self._prim._bodya.ang_vel, self._prim._ra)
        vb: Matrix = self._prim._bodyb.vel + Matrix.cross_product2(
            self._prim._bodyb.ang_vel, self._prim._rb)

        jvb: Matrix = va - vb
        jvb += self._prim._bias
        jvb.negate()

        J: Matrix = self._prim._eff_mass * jvb
        old_impulse: Matrix = self._prim._impulse
        self._prim._impulse += J

        max_impulse: float = dt * self._prim._force_max

        if self._prim._impulse.len_square() > max_impulse * max_impulse:
            self._prim._impulse.normalize()
            self._prim._impulse *= max_impulse

        J = self._prim._impulse - old_impulse
        self._prim._bodya.apply_impulse(J, self._prim._ra)
        self._prim._bodyb.apply_impulse(-J, self._prim._rb)

    def set_value(self, pa: Matrix, pb: Matrix) -> None:
        self._prim._nearest_pa = pa
        self._prim._nearest_pb = pb

    def solve_position(self, dt: float) -> None:
        pass

    def prim(self) -> DistanceConstraintPrimitive:
        return self._prim



import taichi as ti


class Viewport():
    '''the viewport of the render, the origin is bottom left of
        the screen
        '''
    def __init__(self,
                 top_left: ti.Vector = ti.Vector([0.0, 720.0]),
                 bot_right: ti.Vector = ti.Vector([1280.0, 0.0])):

        assert top_left.x < bot_right.x
        assert top_left.y > bot_right.y

        self._top_left: ti.Vector = top_left
        self._bot_right: ti.Vector = bot_right

    @property
    def width(self) -> float:
        return self._bot_right.x - self._top_left.x

    @width.setter
    def width(self, width: float) -> None:
        self._bot_right.x = self._top_left.x + width

    @property
    def height(self) -> float:
        return self._top_left.y - self._bot_right.y

    @height.setter
    def height(self, height: float) -> None:
        self._top_left.y = self._bot_right.y + height

    def set_value(self, width: float, height: float):
        self.width = width
        self.height = height



from typing import TypeVar


class Config():
    SimplexMax: int = 8
    Epsilon: float = 1e-5
    Max: float = 1e37
    PositiveMin: float = 1e-37
    NegativeMin: float = -Max
    Pi: float = 3.14159265
    HalfPi: float = Pi / 2.0
    DoublePi: float = Pi * 2.0
    ReciprocalOfPi: float = 0.3183098861
    GeometryEpsilon: float = 0.00001
    MaxVelocity: float = 1000.0
    MaxAngularVelocity: float = 1000.0
    # render
    BackgroundColor: int = 0x323232
    OuterLineColor: int = 0x00FF00
    FillColor: int = 0x008000
    AxisPointColor: int = 0x00FF00
    AxisLineColor: int = 0x008000
    AABBLineColor: int = 0xFFFF33
    BodyCenterColor: int = 0x660066
    AngleLineXColor: int = 0x0000FF
    AngleLineYColor: int = 0xFF0000
    QueryRectLineColor: int = 0xFF0000
    QueryRaycasFillColor: int = 0x00CCCC
    QueryRaycasOutLineColor: int = 0x33FFFF
    JointPointColor: int = 0xFF0000
    JointLineColor: int = 0x0000FF

    T = TypeVar('T', float, int)

    @staticmethod
    def clamp(num: T, low: T, high: T) -> T:
        assert low <= high

        if num < low:
            return low
        elif num > high:
            return high
        else:
            return num



from __future__ import annotations
from typing import List, Optional

import numpy as np
import taichi as ti

from ..common.config import Config
from ..render.render import Render
from ..math.matrix import Matrix
from ..dynamics.phy_world import PhysicsWorld
from ..dynamics.body import Body
from ..dynamics.constraint.contact import ContactMaintainer
from ..collision.broad_phase.dbvh import DBVH
from ..collision.broad_phase.dbvt import DBVT
from ..collision.broad_phase.aabb import AABB
from ..geometry.shape import ShapePrimitive


class Camera():
    class Viewport():
        '''the viewport of the render, the origin is bottom left of
        the screen
        '''
        def __init__(self,
                     top_left: Matrix = Matrix([0.0, 720.0], 'vec'),
                     bot_right: Matrix = Matrix([1280.0, 0.0], 'vec')):

            assert top_left.x < bot_right.x
            assert top_left.y > bot_right.y

            self._top_left: Matrix = top_left
            self._bot_right: Matrix = bot_right

        @property
        def width(self) -> float:
            return self._bot_right.x - self._top_left.x

        @width.setter
        def width(self, width: float) -> None:
            self._bot_right.x = self._top_left.x + width

        @property
        def height(self) -> float:
            return self._top_left.y - self._bot_right.y

        @height.setter
        def height(self, height: float) -> None:
            self._top_left.y = self._bot_right.y + height

        def set_value(self, width: float, height: float):
            self.width = width
            self.height = height

    def __init__(self):
        self._visible: bool = True
        self._aabb_visible: bool = False
        self._joint_visible: bool = False
        self._body_visible: bool = False
        self._axis_visible: bool = False
        self._dbvh_visible: bool = False
        self._dbvt_visible: bool = False
        self._grid_visible: bool = False
        self._rotation_line_visible: bool = False
        self._center_visible: bool = False
        self._contact_visible: bool = False

        self._meter_to_pixel: float = 33.0
        self._pixel_to_meter: float = 1 / self._meter_to_pixel

        # 1920x1080[80] -> 1280x720[53]
        self._target_meter_to_pixel: float = 53.0
        self._target_pixel_to_meter: float = 1 / self._target_meter_to_pixel

        self._transform: Matrix = Matrix([0.0, 0.0], 'vec')
        self._origin: Matrix = Matrix([0.0, 0.0], 'vec')
        self._viewport = Camera.Viewport()

        self._world: Optional[PhysicsWorld] = None
        self._target_body: Optional[Body] = None
        self._dbvh: Optional[DBVH] = None
        self._dbvt: Optional[DBVT] = None
        self._maintainer: Optional[ContactMaintainer] = None

        self._zoom_factor: float = 1.0
        self._restit: float = 2.0
        self._delta_time: float = 15.0
        self._axis_point_count: int = 20

    # render factory method
    def render(self, gui: ti.GUI) -> None:
        # for i in range(100):
        #     gui.circle([0.5, 0.5], radius=10, color=0x068587)
        # return
        if self.visible:
            # assert self.world is not None
            self.smooth_scale()

            if self.body_visible:
                self.render_body(gui)

            if self.joint_visible:
                self.render_joint(gui)

            if self.axis_visible:
                self.render_axis(gui)

            if self.aabb_visible:
                assert self._dbvt is not None

                for elem in self._dbvt.tree():
                    if elem._body is not None:
                        self.render_aabb(gui, elem._aabb)

            if self.dbvh_visible:
                pass

            if self.dbvt_visible:
                assert self._dbvt is not None
                self.render_dbvt(gui, self._dbvt.root_index())

            if self.grid_visible:
                self.render_grid_scale_line(gui)

            if self.contact_visible:
                self.render_contact(gui)

    @property
    def visible(self) -> bool:
        return self._visible

    @visible.setter
    def visible(self, visible: bool) -> None:
        self._visible = visible

    @property
    def aabb_visible(self) -> bool:
        return self._aabb_visible

    @aabb_visible.setter
    def aabb_visible(self, visible: bool) -> None:
        self._aabb_visible = visible

    @property
    def joint_visible(self) -> bool:
        return self._joint_visible

    @joint_visible.setter
    def joint_visible(self, visible: bool) -> None:
        self._joint_visible = visible

    @property
    def body_visible(self) -> bool:
        return self._body_visible

    @body_visible.setter
    def body_visible(self, visible: bool) -> None:
        self._body_visible = visible

    @property
    def axis_visible(self) -> bool:
        return self._axis_visible

    @axis_visible.setter
    def axis_visible(self, visible: bool) -> None:
        self._axis_visible = visible

    @property
    def dbvh_visible(self) -> bool:
        return self._dbvh_visible

    @dbvh_visible.setter
    def dbvh_visible(self, visible: bool) -> None:
        self._dbvh_visible = visible

    @property
    def dbvt_visible(self) -> bool:
        return self._dbvt_visible

    @dbvt_visible.setter
    def dbvt_visible(self, visible: bool) -> None:
        self._dbvt_visible = visible

    @property
    def grid_visible(self) -> bool:
        return self._grid_visible

    @grid_visible.setter
    def grid_visible(self, visible: bool) -> None:
        self._grid_visible = visible

    @property
    def rot_line_visible(self) -> bool:
        return self._rotation_line_visible

    @rot_line_visible.setter
    def rot_line_visible(self, visible: bool) -> None:
        self._rotation_line_visible = visible

    @property
    def center_visible(self) -> bool:
        return self._center_visible

    @center_visible.setter
    def center_visible(self, visible: bool) -> None:
        self._center_visible = visible

    @property
    def contact_visible(self) -> bool:
        return self._contact_visible

    @contact_visible.setter
    def contact_visible(self, visible: bool) -> None:
        self._contact_visible = visible

    @property
    def meter_to_pixel(self) -> float:
        return self._meter_to_pixel

    @meter_to_pixel.setter
    def meter_to_pixel(self, val: float) -> None:
        # set the target 'meter_to_pixel' value
        # when in render,
        # clamp the scale value
        if val < 1.0:
            self._target_meter_to_pixel = 1.0
            self._target_pixel_to_meter = 1.0
            return

        self._target_meter_to_pixel = val
        self._target_pixel_to_meter = 1.0 / val

    @property
    def transform(self) -> Matrix:
        return self._transform

    @transform.setter
    def transform(self, trans: Matrix) -> None:
        self._transform = trans

    @property
    def world(self) -> PhysicsWorld:
        assert self._world is not None
        return self._world

    @world.setter
    def world(self, world: PhysicsWorld) -> None:
        self._world = world

    @property
    def target_body(self) -> Body:
        assert self._target_body is not None
        return self._target_body

    @target_body.setter
    def target_body(self, body: Body) -> None:
        self._target_body = body

    @property
    def zoom_factor(self) -> float:
        return self._zoom_factor

    @zoom_factor.setter
    def zoom_factor(self, factor: float) -> None:
        self._zoom_factor = factor

    @property
    def viewport(self) -> Viewport:
        return self._viewport

    @viewport.setter
    def viewport(self, viewport: Viewport) -> None:
        self._viewport = viewport

        tmp: List[float] = []
        tmp.append(
            (self._viewport._top_left.x + self._viewport._bot_right.x) * 0.5)
        tmp.append(
            (self._viewport._top_left.y + self._viewport._bot_right.y) * 0.5)
        self._origin.set_value(tmp)

    def smooth_scale(self) -> None:
        # calc the 'meter to pixel' scale according
        # to the 'target meter to pixel' set from
        # the wheel event smooth animation
        inv_dt: float = 1.0 / self._delta_time
        scale: float = self._target_meter_to_pixel - self._meter_to_pixel
        if np.fabs(scale) < 0.1 or self._meter_to_pixel < 1.0:
            self._meter_to_pixel = self._target_meter_to_pixel
        else:
            self._meter_to_pixel -= (1.0 -
                                     np.exp(self._restit * inv_dt)) * scale

        self._pixel_to_meter = 1.0 / self._meter_to_pixel

    def world_to_screen(self, pos: Matrix) -> Matrix:
        orign: Matrix = Matrix([
            self._origin.x + self._transform.x,
            self._origin.y + self._transform.y
        ], 'vec')

        # taichi axis system is radio-based
        view_width: float = self.viewport.width
        view_height: float = self.viewport.height

        tmpx: float = (orign.x + pos.x * self._meter_to_pixel) / view_width
        tmpy: float = (orign.y + pos.y * self._meter_to_pixel) / view_height
        # print(f'({tmpx}, {tmpy})')
        # tmpx = Config.clamp(tmpx, 0.0, 1.0)
        # tmpy = Config.clamp(tmpy, 0.0, 1.0)
        return Matrix([tmpx, tmpy], 'vec')

    def screen_to_world(self, pos: Matrix) -> Matrix:
        orign: Matrix = Matrix([
            self._origin.x + self._transform.x,
            self._origin.y + self._transform.y
        ], 'vec')

        view_width: float = self.viewport.width
        view_height: float = self.viewport.height

        res: Matrix = Matrix([0.0, 0.0], 'vec')
        res.x = pos.x * view_width
        res.y = pos.y * view_height
        res -= orign
        res *= self._pixel_to_meter

        return res

    @property
    def dbvh(self) -> DBVH:
        assert self._dbvh is not None
        return self._dbvh

    @dbvh.setter
    def dbvh(self, dbvh: DBVH) -> None:
        self._dbvh = dbvh

    @property
    def dbvt(self) -> DBVT:
        assert self._dbvt is not None
        return self._dbvt

    @dbvt.setter
    def dbvt(self, dbvt: DBVT):
        self._dbvt = dbvt

    @property
    def delta_time(self) -> float:
        return self._delta_time

    @delta_time.setter
    def delta_time(self, time: float) -> None:
        self._delta_time = time

    @property
    def maintainer(self) -> ContactMaintainer:
        assert self._maintainer is not None
        return self._maintainer

    @maintainer.setter
    def maintainer(self, maintainer: ContactMaintainer) -> None:
        self._maintainer = maintainer

    def render_body(self, gui: ti.GUI) -> None:
        assert self._world is not None

        for bd in self._world._body_list:
            prim: ShapePrimitive = ShapePrimitive()
            prim._shape = bd.shape
            prim._rot = bd.rot
            prim._xform = bd.pos

            Render.rd_shape(gui, prim, self.world_to_screen,
                            self.meter_to_pixel, Config.FillColor)

            if self.center_visible:
                Render.rd_point(gui, self.world_to_screen(prim._xform),
                                Config.BodyCenterColor, 4)

            if self._rotation_line_visible:
                Render.rd_angle_line(gui, prim, self.world_to_screen)

    def render_joint(self, gui: ti.GUI) -> None:
        assert self._world is not None
        for jt in self._world._joint_list:
            if jt.active:
                Render.rd_joint(gui, jt, self.world_to_screen)

    def render_axis(self, gui: ti.GUI) -> None:
        axis_points: List[Matrix] = []

        for i in range(-self._axis_point_count, self._axis_point_count + 1):
            tmp: Matrix = self.world_to_screen(Matrix([0.0, i * 1.0], 'vec'))
            axis_points.append(tmp)
            tmp = self.world_to_screen(Matrix([i * 1.0, 0.0], 'vec'))
            axis_points.append(tmp)

        Render.rd_points(gui, axis_points, Config.AxisPointColor)
        Render.rd_line(gui,
                       axis_points[0],
                       axis_points[-2],
                       color=Config.AxisLineColor)
        Render.rd_line(gui,
                       axis_points[1],
                       axis_points[-1],
                       color=Config.AxisLineColor)

    def render_aabb(self, gui: ti.GUI, aabb: AABB) -> None:
        tmp1: Matrix = self.world_to_screen(aabb.top_left)
        tmp2: Matrix = self.world_to_screen(aabb.bot_right)
        Render.rd_rect(gui, tmp1, tmp2, Config.AABBLineColor)

    def render_dbvt(self, gui: ti.GUI, node_idx: int) -> None:
        if node_idx == -1:
            return

        assert self._dbvt is not None
        self.render_dbvt(gui, self._dbvt.tree()[node_idx]._left_idx)
        self.render_dbvt(gui, self._dbvt.tree()[node_idx]._right_idx)

        aabb: AABB = self._dbvt.tree()[node_idx]._aabb
        if not self._dbvt.tree()[node_idx].is_leaf():
            self.render_aabb(gui, aabb)

    def render_contact(self, gui: ti.GUI) -> None:
        pass

    def render_grid_scale_line(self, gui: ti.GUI) -> None:
        pass

    def render_dbvh(self, gui: ti.GUI, node: DBVH.Node) -> None:
        pass



from abc import ABC, abstractmethod
from enum import IntEnum, unique

import numpy as np


@unique
class JointType(IntEnum):
    BASE = -1
    Distance = 0
    Point = 1
    Rotation = 2
    Orientation = 3
    Pulley = 4
    Prismatic = 5
    Wheel = 6
    Revolute = 7


class Joint(ABC):
    def __init__(self):
        self._active: bool = True
        self._type: JointType = JointType.BASE
        self._id: int = 0

    @abstractmethod
    def prepare(self, dt: float) -> None:
        raise NotImplementedError

    @abstractmethod
    def solve_velocity(self, dt: float) -> None:
        raise NotImplementedError

    @abstractmethod
    def solve_position(self, dt: float) -> None:
        raise NotImplementedError

    @property
    def active(self) -> bool:
        return self._active

    @active.setter
    def active(self, active: bool) -> None:
        self._active = active

    def type(self) -> JointType:
        return self._type

    @property
    def id(self) -> int:
        return self._id

    @id.setter
    def id(self, val: int) -> None:
        self._id = val

    @staticmethod
    def natural_frequency(freq: float) -> float:
        return 2 * np.pi * freq

    @staticmethod
    def spring_damping_cofficient(mass: float, natural_freq: float,
                                  damping_radio: float) -> float:
        return damping_radio * 2.0 * mass * natural_freq

    @staticmethod
    def spring_stiff(mass: float, natural_freq: float) -> float:
        return mass * natural_freq * natural_freq

    @staticmethod
    def constraint_impulse_mixing(dt: float, stiff: float,
                                  damping: float) -> float:
        cim: float = dt * (dt * stiff + damping)
        return 0.0 if np.isclose(cim, 0.0) else 1.0 / cim

    @staticmethod
    def error_reduction_parameter(dt: float, stiff: float,
                                  damping: float) -> float:
        erp: float = dt * stiff + damping
        return 0.0 if np.isclose(erp, 0.0) else stiff / erp



from typing import List

from ...math.matrix import Matrix
from ..body import Body
from .joint import Joint, JointType


class PointJointPrimitive():
    def __init__(self):
        self._bodya: Body = Body()
        self._local_pointa: Matrix = Matrix([0.0, 0.0], 'vec')
        self._target_point: Matrix = Matrix([0.0, 0.0], 'vec')
        self._normal: Matrix = Matrix([0.0, 0.0], 'vec')

        self._damping: float = 0.0
        self._stiff: float = 0.0
        self._freq: float = 8.0
        self._force_max: float = 1000.0
        self._damping_radio: float = 1.0
        self._gamma: float = 0.0
        self._bias: Matrix = Matrix([0.0, 0.0], 'vec')
        self._eff_mass: Matrix = Matrix([0.0, 0.0, 0.0, 0.0])
        self._impulse: Matrix = Matrix([0.0, 0.0], 'vec')


class PointJoint(Joint):
    def __init__(self, prim: PointJointPrimitive = PointJointPrimitive()):
        super().__init__()
        self._type: JointType = JointType.Point
        self._prim: PointJointPrimitive = prim
        self._factor: float = 0.22

    def set_value(self, prim: PointJointPrimitive) -> None:
        self._prim = prim

    def prepare(self, dt: float) -> None:
        if self._prim._bodya is None:
            return

        bodya: Body = self._prim._bodya
        m_a: float = bodya.mass
        im_a: float = bodya.inv_mass
        ii_a: float = bodya.inv_inertia

        if self._prim._freq > 0.0:
            nf: float = Joint.natural_frequency(self._prim._freq)
            self._prim._stiff = Joint.spring_stiff(m_a, nf)
            self._prim._damping = Joint.spring_damping_cofficient(
                m_a, nf, self._prim._damping_radio)
        else:
            self._prim._stiff = 0.0
            self._prim._damping = 0.0

        self._prim._gamma = Joint.constraint_impulse_mixing(
            dt, self._prim._stiff, self._prim._damping)
        erp: float = Joint.error_reduction_parameter(dt, self._prim._stiff,
                                                     self._prim._damping)

        pa: Matrix = bodya.to_world_point(self._prim._local_pointa)
        ra: Matrix = pa - bodya.pos
        pb: Matrix = self._prim._target_point

        self._prim._bias = (pa - pb) * erp
        k: Matrix = Matrix([0.0, 0.0, 0.0, 0.0])
        data_arr: List[float] = []
        data_arr.append(im_a + ra.y * ra.y * ii_a)
        data_arr.append(-ra.x * ra.y * ii_a)
        data_arr.append(data_arr[1])
        data_arr.append(im_a + ra.x * ra.x * ii_a)

        data_arr[0] += self._prim._gamma
        data_arr[3] += self._prim._gamma

        k.set_value(data_arr)
        self._prim._eff_mass = k.invert()
        bodya.apply_impulse(self._prim._impulse, ra)

    def solve_velocity(self, dt: float) -> None:
        if self._prim._bodya is None:
            return

        ra: Matrix = self._prim._bodya.to_world_point(
            self._prim._local_pointa) - self._prim._bodya.pos
        va: Matrix = self._prim._bodya.vel + Matrix.cross_product2(
            self._prim._bodya.ang_vel, ra)

        jvb: Matrix = va
        jvb += self._prim._bias
        jvb += self._prim._impulse * self._prim._gamma
        jvb.negate()

        J: Matrix = self._prim._eff_mass * jvb
        old_impulse: Matrix = self._prim._impulse
        self._prim._impulse += J

        max_impulse: float = dt * self._prim._force_max
        if self._prim._impulse.len_square() > max_impulse * max_impulse:
            self._prim._impulse.normalize()
            self._prim._impulse *= max_impulse

        J = self._prim._impulse - old_impulse
        self._prim._bodya.apply_impulse(J, ra)

    def solve_position(self, dt: float) -> None:
        pass

    def prim(self) -> PointJointPrimitive:
        return self._prim



from typing import Optional, Union, List, cast

import taichi as ti

from ..dynamics.body import Body
from .joint.joint import Joint
from ..common.random import RandomGenerator
from ..geometry.shape import Capsule, Circle, Edge, Polygon, Shape


@ti.data_oriented
class PhysicsWorld():
    def __init__(self, body_len: int = 60):
        # env var
        self._grav_ena: bool = True
        self._grav: ti.Vector = ti.Vector([0.0, -1.0])

        self._damping_ena: bool = True
        self._linear_vel_damping: float = 0.9
        self._ang_vel_damping: float = 0.9

        self._body_list: List[Body] = []
        self._joint_list: List[Joint] = []

        # body physics params
        self._body_len: int = body_len

        # rot line
        self._rot_line_hst = ti.Vector.field(2, float, shape=self._body_len)
        self._rot_line_hed = ti.Vector.field(2, float, shape=self._body_len)
        self._rot_line_vst = ti.Vector.field(2, float, shape=self._body_len)
        self._rot_line_ved = ti.Vector.field(2, float, shape=self._body_len)

        self._mass = ti.field(float, shape=self._body_len)
        self._inertia = ti.field(float, shape=self._body_len)
        # 1: kinematic 2: static 3: dynamic
        self._phy_type = ti.field(int, shape=self._body_len)
        # 1: circle 2: Edge 3: triangle 4: rect 5: pentagon
        # 6: hexagon 7: capsule
        self._shape_type = ti.field(int, shape=self._body_len)
        # circle
        self._cirpos = ti.Vector.field(2, float, shape=self._body_len)
        self._cirspos = ti.Vector.field(2, float, shape=self._body_len)
        self._cir_rad = ti.field(float, shape=self._body_len)
        self._scir_rad = ti.field(float, shape=self._body_len)
        # edge
        self._edgpos = ti.Vector.field(2, float, shape=self._body_len)
        self._edg_st = ti.Vector.field(2, float, shape=self._body_len)
        self._edg_sst = ti.Vector.field(2, float, shape=self._body_len)
        self._edg_ed = ti.Vector.field(2, float, shape=self._body_len)
        self._edg_sed = ti.Vector.field(2, float, shape=self._body_len)

        # polygon triangle
        self._poly_tripos = ti.Vector.field(2, float, shape=self._body_len)
        self._poly_trispos = ti.Vector.field(2, float, shape=self._body_len)
        self._poly_trist = ti.Vector.field(2, float, shape=(self._body_len, 3))
        self._poly_tried = ti.Vector.field(2, float, shape=(self._body_len, 3))
        self._poly_trisst = ti.Vector.field(2,
                                            float,
                                            shape=(self._body_len, 3))
        self._poly_trised = ti.Vector.field(2,
                                            float,
                                            shape=(self._body_len, 3))
        self._poly_tri_a = ti.Vector.field(2, float, shape=(self._body_len, 1))
        self._poly_tri_b = ti.Vector.field(2, float, shape=(self._body_len, 1))
        self._poly_tri_c = ti.Vector.field(2, float, shape=(self._body_len, 1))

        # polygon rect
        self._poly_recpos = ti.Vector.field(2, float, shape=self._body_len)
        self._poly_recspos = ti.Vector.field(2, float, shape=self._body_len)
        self._poly_recst = ti.Vector.field(2, float, shape=(self._body_len, 4))
        self._poly_reced = ti.Vector.field(2, float, shape=(self._body_len, 4))
        self._poly_recsst = ti.Vector.field(2,
                                            float,
                                            shape=(self._body_len, 4))
        self._poly_recsed = ti.Vector.field(2,
                                            float,
                                            shape=(self._body_len, 4))

        self._poly_rec_a = ti.Vector.field(2, float, shape=(self._body_len, 2))
        self._poly_rec_b = ti.Vector.field(2, float, shape=(self._body_len, 2))
        self._poly_rec_c = ti.Vector.field(2, float, shape=(self._body_len, 2))

        # polygon pen
        self._poly_penpos = ti.Vector.field(2, float, shape=self._body_len)
        self._poly_penspos = ti.Vector.field(2, float, shape=self._body_len)
        self._poly_penst = ti.Vector.field(2, float, shape=(self._body_len, 5))
        self._poly_pened = ti.Vector.field(2, float, shape=(self._body_len, 5))
        self._poly_pensst = ti.Vector.field(2,
                                            float,
                                            shape=(self._body_len, 5))
        self._poly_pensed = ti.Vector.field(2,
                                            float,
                                            shape=(self._body_len, 5))

        self._poly_pen_a = ti.Vector.field(2, float, shape=(self._body_len, 3))
        self._poly_pen_b = ti.Vector.field(2, float, shape=(self._body_len, 3))
        self._poly_pen_c = ti.Vector.field(2, float, shape=(self._body_len, 3))

        # capsule
        self._cap_pos = ti.Vector.field(2, float, shape=self._body_len)
        self._cap_spos = ti.Vector.field(2, float, shape=self._body_len)
        self._cap_width = ti.field(float, shape=self._body_len)
        self._cap_height = ti.field(float, shape=self._body_len)
        self._cap_c1 = ti.Vector.field(2, float, shape=self._body_len)
        self._cap_c2 = ti.Vector.field(2, float, shape=self._body_len)
        self._cap_p = ti.Vector.field(2, float, shape=(self._body_len, 4))
        self._cap_rec_a = ti.Vector.field(2, float, shape=(self._body_len, 2))
        self._cap_rec_b = ti.Vector.field(2, float, shape=(self._body_len, 2))
        self._cap_rec_c = ti.Vector.field(2, float, shape=(self._body_len, 2))

        self._vel = ti.Vector.field(2, float, shape=self._body_len)
        self._rot = ti.field(float, shape=self._body_len)
        self._ang_vel = ti.field(float, shape=self._body_len)
        self._force = ti.Vector.field(2, float, shape=self._body_len)
        self._torque = ti.field(float, shape=self._body_len)

    def create_body(self):
        body: Body = Body()
        body.id = RandomGenerator.unique()
        self._body_list.append(body)
        return body

    def init_data(self):
        bd_len: int = len(self._body_list)
        for i in range(bd_len):
            self._mass[i] = self._body_list[i].mass
            self._inertia[i] = self._body_list[i].inertia
            self._phy_type[i] = self._body_list[i].type + 1
            self._vel[i] = ti.Vector(
                [self._body_list[i].vel.x, self._body_list[i].vel.y])
            self._rot[i] = self._body_list[i].rot
            self._ang_vel[i] = self._body_list[i].ang_vel
            self._force[i] = ti.Vector(
                [self._body_list[i].forces.x, self._body_list[i].forces.y])
            self._torque[i] = self._body_list[i].torques

            pos = ti.Vector(
                [self._body_list[i].pos.x, self._body_list[i].pos.y])
            if self._body_list[i].shape.type == Shape.Type.Circle:
                cir: Circle = cast(Circle, self._body_list[i].shape)
                self._shape_type[i] = 1
                self._cirpos[i] = pos
                self._cir_rad[i] = cir.radius

            elif self._body_list[i].shape.type == Shape.Type.Edge:
                edg: Edge = cast(Edge, self._body_list[i].shape)
                self._shape_type[i] = 2
                self._edg_st[i] = ti.Vector([edg.start.x, edg.start.y])
                self._edg_ed[i] = ti.Vector([edg.end.x, edg.end.y])

            elif self._body_list[i].shape.type == Shape.Type.Polygon:
                poly: Polygon = cast(Polygon, self._body_list[i].shape)
                vert_len: int = len(poly.vertices)
                assert vert_len >= 4

                if vert_len == 4:
                    self._shape_type[i] = 3
                    self._poly_tripos[i] = pos
                    for j in range(3):
                        self._poly_trist[i, j] = ti.Vector(
                            [poly.vertices[j].x, poly.vertices[j].y])
                        self._poly_tried[i, j] = ti.Vector(
                            [poly.vertices[j + 1].x, poly.vertices[j + 1].y])

                elif vert_len == 5:
                    self._shape_type[i] = 4
                    self._poly_recpos[i] = pos
                    for j in range(4):
                        self._poly_recst[i, j] = ti.Vector(
                            [poly.vertices[j].x, poly.vertices[j].y])
                        self._poly_reced[i, j] = ti.Vector(
                            [poly.vertices[j + 1].x, poly.vertices[j + 1].y])

                elif vert_len == 6:
                    self._shape_type[i] = 5
                    self._poly_penpos[i] = pos
                    for j in range(5):
                        self._poly_penst[i, j] = ti.Vector(
                            [poly.vertices[j].x, poly.vertices[j].y])
                        self._poly_pened[i, j] = ti.Vector(
                            [poly.vertices[j + 1].x, poly.vertices[j + 1].y])

                elif vert_len == 7:
                    self._shape_type[i] = 6

            elif self._body_list[i].shape.type == Shape.Type.Capsule:
                self._shape_type[i] = 7
                self._cap_pos[i] = pos
                cap: Capsule = cast(Capsule, self._body_list[i].shape)
                self._cap_width[i] = cap.width
                self._cap_height[i] = cap.height

    @ti.kernel
    def step_velocity(self, dt: float):
        lvd = 1.0 / (1.0 + dt * self._linear_vel_damping)
        avd = 1.0 / (1.0 + dt * self._ang_vel_damping)
        g = self._grav

        for i in range(self._body_len):
            if self._phy_type[i] == -1:
                self._vel[i] = ti.Vector([0.0, 0.0])
                self._ang_vel[i] = 0.0

            elif self._phy_type[i] == 0:
                self._force[i] = self._force[i] + self._mass[i] * g
                self._vel[
                    i] = self._vel[i] + self._force[i] / self._mass[i] * dt
                self._ang_vel[i] = self._ang_vel[
                    i] + self._torque[i] / self._inertia[i] * dt

                self._vel[i] = self._vel[i] * lvd
                self._ang_vel[i] = self._ang_vel[i] * avd

            elif self._phy_type[i] == 1:
                self._vel[
                    i] = self._vel[i] + self._force[i] / self._mass[i] * dt
                self._ang_vel[i] = self._ang_vel[
                    i] + self._torque[i] / self._inertia[i] * dt

                self._vel[i] = self._vel[i] * lvd
                self._ang_vel[i] = self._ang_vel[i] * avd

    @ti.kernel
    def step_position(self, dt: float):
        for i in range(self._body_len):
            if self._phy_type[i] == 0:
                pass

            elif self._phy_type[i] == 1 or self._phy_type[i] == 3:
                # self._pos[i] = self._pos[i] + self._vel[i] * dt
                self._rot[i] = self._rot[i] + self._ang_vel[i] * dt
                self._force[i] = ti.Vector([0.0, 0.0])
                self._torque[i] = 0.0

    def clear_all_bodies(self) -> None:
        self._body_list.clear()

    def clear_all_joints(self) -> None:
        self._joint_list.clear()



from typing import List, Dict, Optional, Tuple

import numpy as np

from ...math.matrix import Matrix
from ...common.config import Config
from ...dynamics.body import Body
from ..broad_phase.aabb import AABB


class DBVH():
    '''Dynamic Bounding Volume Hierarchy
    This is implemented by traditional binary search tree
    '''
    class Node():
        def __init__(self,
                     body: Optional[Body] = None,
                     aabb: Optional[AABB] = None):
            self._parent: Optional[DBVH.Node] = None
            self._left: Optional[DBVH.Node] = None
            self._right: Optional[DBVH.Node] = None
            self._body: Optional[Body] = body
            self._aabb: Optional[AABB] = aabb

        def separate(self, node):
            '''check if the node is the child of self,
            if true, separate the parent-child

            Parameters
            ----------
            node : Node
                other node
            '''
            if node == None:
                return

            if node.is_leaf() and node.is_root():
                return

            if node == self._left:
                self._left = None
                node._parent = None
                if self._right != None:
                    self._aabb = self._right._aabb
            elif node == self._right:
                self._right = None
                node._parent = None
                if self._left != None:
                    self._aabb = self._left._aabb

        def swap(self, src, target):
            '''swap the self's child src node with target node

            Parameters
            ----------
            src : Node
                self's child node
            target : Node
                swap node
            '''
            if src == self._left:
                self.separate(self._left)
                target._parent = self
                self._left = target

            elif src == self._right:
                self.separate(self._right)
                target._parent = self
                self._right = target

        def is_leaf(self) -> bool:
            return self._left == None and self._right == None

        def is_branch(self) -> bool:
            return self._left != None and self._right != None and self._parent != None

        def is_root(self) -> bool:
            return self._parent == None

        def clear(self) -> bool:
            self._body = None
            if self._aabb != None:
                self._aabb.clear()

    def __init__(self):
        self._root: Optional[DBVH.Node] = None
        self._profile: float = 0.0
        self._leaf_factor: float = 0.5
        self._leaves: Dict[Body, DBVH.Node] = {}

    def find_helper(self, body: Body) -> bool:
        for b in self._leaves.keys():
            if b == body:
                return True
        return False

    def insert(self, body: Body):
        def _get_cost(aabb: AABB) -> DBVH.Node:
            low_cost_node: Optional[DBVH.Node] = None
            cost_value: float = Config.Max

            for val in self._leaves.values():
                assert val != None
                cost: float = self._total_cost(val, aabb)

                if cost < cost_value:
                    cost_value = cost
                    low_cost_node = val

            return low_cost_node

        if not self.find_helper(body):
            return

        aabb: AABB = AABB.from_body(body)
        aabb.expand(self._leaf_factor)

        if self._root == None:
            self._root = DBVH.Node(body, aabb)
            self._leaves[body] = self._root
            return

        if self._root.is_leaf() and self._root.is_root():
            self._merge(self._root, aabb, body)
            self._update(self._root)
            return

        target = _get_cost(aabb)
        self._merge(target, aabb, body)
        self._balance(self._root)
        self._update(target)
        for val in self._leaves.values():
            self._update(val)

    def update(self, body: Body):
        assert body != None

        if not self.find_helper(body):
            return

        thin: AABB = AABB.from_body(body)
        thin.expand(0.1)
        if not thin.is_subset(self._leaves[body]._aabb):
            node: Optional[DBVH.Node] = self.extract(body)
            self._insert(node)

    def extract(self, body: Body):
        def _move_up(branch: DBVH.Node) -> DBVH.Node:
            parent: DBVH.Node = branch._parent
            child: DBVH.Node = branch._left if branch._left != None else branch._right

            if branch.is_root():
                child._parent = None
                self._root = child
                return child

            parent.swap(branch, child)
            branch._parent = None
            branch._aabb.clear()
            return child

        if not self.find_helper(body):
            return None

        source: DBVH.Node = self._leaves[body]
        parent: DBVH.Node = source._parent
        parent.separate(source)

        if source.is_root() and source.is_leaf() and len(self._leaves) < 2:
            self._root = None
            return source

        child: DBVH.Node = _move_up(parent)
        self._balance(child._parent)
        self._update(child._parent)
        return source

    def erase(self, body: Body):
        target: Optional[DBVH.Node] = self.extract(body)
        if target == None:
            return

        target._body = None
        target._aabb.clear()
        del self._leaves[body]

    def clean_up(self, node: Optional[Node]):
        if node == None:
            return

        self.clean_up(node._left)
        self.clean_up(node._right)
        node = None

    def root(self) -> Optional[Node]:
        return self._root

    def raycast(self, start: Matrix, dirn: Matrix) -> List[Body]:
        res: List[Body] = []
        self._raycast(res, self._root, start, dirn)
        return res

    def generate(self) -> List[Tuple[Body, Body]]:
        self._profile = 0.0
        res: List[Tuple[Body, Body]] = []
        self._generate(self._root, res)
        return res

    def leaves(self) -> Dict[Body, Node]:
        return self._leaves

    def query(self, src: AABB, nodes: List[Node], skip_body: Body = None):
        DBVH.query_nodes(self._root, src, nodes, skip_body)

    @staticmethod
    def query_nodes(node: Node,
                    aabb: AABB,
                    nodes: List[Node],
                    skip_body: Body = None):
        if node == None or not aabb.collide(node._aabb):
            return

        if skip_body != None and node._body == skip_body:
            return

        if node.is_branch() or node.is_root():
            DBVH.query_nodes(node._left, aabb, nodes, skip_body)
            DBVH.query_nodes(node._right, aabb, nodes, skip_body)
            return

        if node.is_leaf():
            nodes.append(node)

    def _raycast(self, res: List[Body], node: Optional[Node], start: Matrix,
                 dirn: Matrix):
        if node == None:
            return

        if node._aabb.raycast(start, dirn):
            if node.is_leaf():
                res.append(node._body)
            else:
                self._raycast(res, node._left, start, dirn)
                self._raycast(res, node._right, start, dirn)

    def _insert(self, node: Optional[Node]):
        def _get_cost(target: DBVH.Node, aabb: AABB) -> DBVH.Node:
            low_cost_node: Optional[DBVH.Node] = None
            cost_value: float = Config.Max

            for val in self._leaves.values():
                assert val != None
                if val == target:
                    continue

                cost: float = self._total_cost(val, aabb)

                if cost < cost_value:
                    cost_value = cost
                    low_cost_node = val

            return low_cost_node

        if node == None:
            return

        aabb: AABB = AABB.from_body(node._body)
        aabb.expand(self._leaf_factor)
        node._aabb = aabb

        if self._root == None:
            self._root = node
            return

        if self._root.is_leaf() and self._root.is_root():
            self._merge2(self._root, node)
            self._update(self._root)
            return

        if self._root.is_root():
            target: Optional[DBVH.Node] = _get_cost(node, aabb)
            self._merge2(target, node)
            self._balance(self._root)

            for val in self._leaves.values():
                if val == node:
                    continue
                self._update(val)

    def _delta_cost(self, node: Optional[Node], aabb: AABB) -> float:
        if node == None:
            return 0.0

        if node.is_leaf():
            return AABB.unite(node._aabb, aabb).surface_area()

        return AABB.unite(node._aabb,
                          aabb).surface_area() - node._aabb.surface_area()

    def _total_cost(self, node: Optional[Node], aabb: AABB) -> float:
        if node == None:
            return None

        cost: float = 0.0
        tmp: float = self._delta_cost(node, aabb)
        cost += tmp
        return cost + self._total_cost(node._parent, aabb)

    def _merge(self, node: Optional[Node], aabb: AABB, body: Body) -> Node:
        assert node != None

        new_node: DBVH.Node = DBVH.Node(body, aabb)
        copy: DBVH.Node = DBVH.Node(node._body, node._aabb)

        self._leaves[new_node._body] = new_node
        if node.is_leaf():
            self._leaves[node._body] = copy

        node._body = None
        node._aabb = AABB.unite(aabb, node._aabb)
        node._left = copy
        node._right = new_node
        copy._parent = node
        new_node._parent = node

        return new_node

    def _merge2(self, target: Optional[Node], src: Optional[Node]):
        assert target != None and src != None
        assert src.is_leaf()

        copy: DBVH.Node = DBVH.Node(target._body, target._aabb)
        if target.is_leaf():
            self._leaves[target._body] = copy

        target._body = None
        target._aabb = AABB.unite(copy._aabb, src._aabb)
        target._left = copy
        target._right = src
        copy._parent = target
        src._parent = target

    def _update(self, parent: Optional[Node]):
        if parent == None:
            return

        if parent.is_branch() or parent.is_root():
            parent._aabb = AABB.unite(parent._left._aabb, parent._right._aabb)

        self._update(parent._parent)

    def _balance(self, node: Optional[Node]):
        def _LL(node: Optional[DBVH.Node]):
            if node == None or node.is_root():
                return

            if node._parent == self._root:
                parent = node._parent
                right = node._right
                parent._left = right
                node._parent = None

                node._right = parent
                right._parent = parent
                parent._parent = node
                self._root = node
                return

            parent = node._parent
            right = node._right
            garnd_parent = parent._parent
            node._parent = garnd_parent

            if parent == garnd_parent._left:
                garnd_parent._left = node
            else:
                garnd_parent._right = node

            node._right = parent
            parent._parent = node
            parent._left = right
            right._parent = parent

        def _RR(node: Optional[DBVH.Node]):
            if node == None or node.is_root():
                return

            if node._parent == self._root:
                parent = node._parent
                left = node._left
                parent._right = left
                node._parent = None

                node._left = parent
                left._parent = parent
                parent._parent = node
                self._root = node
                return

            parent = node._parent
            left = node._left
            garnd_parent = parent._parent

            node._parent = garnd_parent
            if parent == garnd_parent._right:
                garnd_parent._right = node
            else:
                garnd_parent._left = node

            node._left = parent
            parent._parent = node

            parent._right = left
            left._parent = parent

        if node == None:
            return

        left_height: int = self._height(node._left)
        right_height: int = self._height(node._right)
        if np.fabs(left_height - right_height) <= 1:
            return

        # left unbalance
        if left_height > right_height:
            ll_height: int = self._height(node._left._left)
            lr_height: int = self._height(node._left._right)

            if ll_height < lr_height:
                _RR(node._left._right)
            else:
                _LL(node._left._left)

            _LL(node._left)

        # right unbalance
        else:
            rr_height: int = self._height(node._right._right)
            rl_height: int = self._height(node._right._left)
            if rr_height < rl_height:
                _LL(node._right._left)
            else:
                _RR(node._right._right)
            _RR(node._right)

        self._balance(node._left)
        self._balance(node._right)
        self._balance(node._parent)

    # check if children collide with each other
    def _generate(self, node: Optional[Node], pairs: List[Tuple[Body, Body]]):
        if node == None or node.is_leaf():
            return

        res: bool = AABB.unite(node._left._aabb, node._right._aabb)
        self._profile += 1

        if res:
            self._generate2(node._left, node._right, pairs)

        self._generate(node._left, pairs)
        self._generate(node._right, pairs)

    def _generate2(self, left: Optional[Node], right: Optional[Node],
                   pairs: List[Tuple[Body, Body]]):
        if left == None or right == None:
            return

        res: bool = left._aabb.collide(right._aabb) or left._aabb.is_subset(
            right._aabb)
        self._profile += 1

        if not res:
            return

        if left.is_leaf() and right.is_leaf():
            if left._body.bitmask & right._body.bitmask:
                if AABB.from_body(left._body).collide(
                        AABB.from_body(right._body)):
                    pairs.append((left._body, right._body))

        if left.is_leaf() and right.is_branch():
            self._generate2(left, right._left, pairs)
            self._generate2(left, right._right, pairs)

        if right.is_leaf() and left.is_branch():
            self._generate2(right, left._left, pairs)
            self._generate2(right, left._right, pairs)

        if left.is_branch() and right.is_branch():
            self._generate2(left._left, right, pairs)
            self._generate2(left._right, right, pairs)

    def _height(self, node: Optional[Node]) -> int:
        return 0 if node == None else np.fmax(self._height(node._left),
                                              self._height(node._right)) + 1



from typing import List, Dict, cast

import numpy as np

from ...math.matrix import Matrix
from ...common.config import Config
from ...collision.algorithm.gjk import PointPair
from ...collision.detector import Collsion
from ..body import Body


# FIXME:
def generate_relation(bodya: Body, bodyb: Body) -> int:
    # Combine two 32-bit id into one 64-bit id in unique form
    ida: int = bodya.id
    idb: int = bodyb.id

    return ida + idb


class VelocityConstraintPoint():
    def __init__(self):
        self._ra: Matrix = Matrix([0.0, 0.0], 'vec')
        self._rb: Matrix = Matrix([0.0, 0.0], 'vec')
        self._va: Matrix = Matrix([0.0, 0.0], 'vec')
        self._vb: Matrix = Matrix([0.0, 0.0], 'vec')
        self._normal: Matrix = Matrix([0.0, 0.0], 'vec')
        self._tangent: Matrix = Matrix([0.0, 0.0], 'vec')
        self._vel_bias: Matrix = Matrix([0.0, 0.0], 'vec')
        self._bias: float = 0.0
        self._penetration: float = 0.0
        self._restit: float = 0.8
        self._eff_mass_normal: float = 0.0
        self._eff_mass_tangent: float = 0.0
        self._accum_normal_impulse: float = 0.0
        self._accum_tangent_impulse: float = 0.0


class ContactConstraintPoint():
    def __init__(self):
        self._relation: int = 0
        self._fric: float = 0.2
        self._active: bool = True
        self._locala: Matrix = Matrix([0.0, 0.0], 'vec')
        self._localb: Matrix = Matrix([0.0, 0.0], 'vec')
        self._bodya: Body = Body()
        self._bodyb: Body = Body()
        self._vcp: VelocityConstraintPoint = VelocityConstraintPoint()


class ContactMaintainer():
    def __init__(self):
        self._penetration_max: float = 0.01
        self._bias_factor: float = 0.03
        self._contact_table: Dict[int, List[ContactConstraintPoint]] = {}

    def clear_all(self) -> None:
        self._contact_table.clear()

    def solve_velocity(self, dt: float) -> None:
        for val in self._contact_table.values():
            if len(val) == 0 or not val[0]._active:
                continue

            for ccp in val:
                vcp: VelocityConstraintPoint = ccp._vcp
                wa: Matrix = Matrix.cross_product2(ccp._bodya.ang_vel, vcp._ra)
                wb: Matrix = Matrix.cross_product2(ccp._bodyb.ang_vel, vcp._rb)
                vcp._va = ccp._bodya.vel + wa
                vcp._vb = ccp._bodyb.vel + wb

                dv: Matrix = vcp._va - vcp._vb
                jv: float = -1.0 * vcp._normal.dot(dv - vcp._vel_bias)
                lambda_n: float = vcp._eff_mass_normal * jv
                old_impulse: float = vcp._accum_normal_impulse
                vcp._accum_normal_impulse = np.fmax(old_impulse + lambda_n, 0)

                lambda_n = vcp._accum_normal_impulse - old_impulse
                impulse_n: Matrix = vcp._normal * lambda_n

                ccp._bodya.apply_impulse(impulse_n, vcp._ra)
                ccp._bodyb.apply_impulse(-impulse_n, vcp._rb)

                vcp._va = ccp._bodya.vel + Matrix.cross_product2(
                    ccp._bodya.ang_vel, vcp._ra)
                vcp._vb = ccp._bodyb.vel + Matrix.cross_product2(
                    ccp._bodyb.ang_vel, vcp._rb)
                dv = vcp._va - vcp._vb

                jvt: float = vcp._tangent.dot(dv)
                lambda_t: float = vcp._eff_mass_tangent * -jvt

                maxT: float = ccp._fric * vcp._accum_normal_impulse
                old_impulse = vcp._accum_tangent_impulse

                vcp._accum_tangent_impulse = Config.clamp(
                    old_impulse + lambda_t, -maxT, maxT)
                lambda_t = vcp._accum_tangent_impulse - old_impulse
                impulse_t: Matrix = vcp._tangent * lambda_t

                ccp._bodya.apply_impulse(impulse_t, vcp._ra)
                ccp._bodyb.apply_impulse(-impulse_t, vcp._rb)

    def solve_position(self, dt: float) -> None:
        for val in self._contact_table.values():
            if len(val) == 0 or not val[0]._active:
                continue

            for ccp in val:
                vcp: VelocityConstraintPoint = ccp._vcp
                bodya: Body = ccp._bodya
                bodyb: Body = ccp._bodyb
                pa: Matrix = vcp._ra + bodya.pos
                pb: Matrix = vcp._rb + bodyb.pos
                c: Matrix = pa - pb

                # already solved by vel
                if c.dot(vcp._normal) < 0.0:
                    continue

                bias: float = self._bias_factor * np.fmax(
                    c.len() - self._penetration_max, 0.0)
                val_lambda: float = vcp._eff_mass_normal * bias
                impulse: Matrix = vcp._normal * val_lambda

                if bodya.type != Body.Type.Static and not ccp._bodya.sleep:
                    bodya.pos += impulse * bodya.inv_mass
                    bodya.rot += bodya.inv_inertia * vcp._ra.cross(impulse)

                if bodyb.type != Body.Type.Static and not ccp._bodyb.sleep:
                    bodyb.pos -= impulse * bodyb.inv_mass
                    bodyb.rot -= bodyb.inv_inertia * vcp._rb.cross(impulse)

    def add(self, collision: Collsion) -> None:
        assert collision._bodya is not None
        assert collision._bodyb is not None
        bodya: Body = collision._bodya
        bodyb: Body = collision._bodyb

        # print(f'bodya id: {bodya.id}')
        # print(f'bodyb id: {bodyb.id}')
        relation: int = generate_relation(bodya, bodyb)
        # print(f'relation: {relation}')

        contact_list: List[ContactConstraintPoint] = []
        if self._contact_table.get(relation, False):
            # print('sssss')
            contact_list = self._contact_table[relation]
        else:
            self._contact_table[relation] = []
            # print('zzzzzz')
            contact_list = self._contact_table[relation]

        for elem in collision._contact_list:
            # print(f'pa: ({elem._pa.x},{elem._pa.y})')
            # print(f'pb: ({elem._pb.x},{elem._pb.y})')

            existed: bool = False
            locala: Matrix = bodya.to_local_point(elem._pa)
            localb: Matrix = bodyb.to_local_point(elem._pb)

            for contact in contact_list:
                # print('x')
                is_pointa: bool = np.isclose(contact._locala._val,
                                             locala._val).all()
                is_pointb: bool = np.isclose(contact._localb._val,
                                             localb._val).all()

                if is_pointa and is_pointb:
                    # satisfy the condition, transmit the old
                    # accumulated value to new value
                    contact._locala = locala
                    contact._localb = localb
                    self.prepare(contact, elem, collision)
                    existed = True
                    break

            if existed:
                continue

            # no eligible contact, push new contact points
            ccp: ContactConstraintPoint = ContactConstraintPoint()
            ccp._locala = locala
            ccp._localb = localb
            ccp._relation = relation
            self.prepare(ccp, elem, collision)
            contact_list.append(ccp)
            # print(f'ct len: {len(self._contact_table[relation])}')

    def prepare(self, ccp: ContactConstraintPoint, pair: PointPair,
                collision: Collsion) -> None:
        assert collision._bodya is not None
        assert collision._bodyb is not None

        # NOTE: return val by ccp params
        ccp._bodya = collision._bodya
        ccp._bodyb = collision._bodyb
        ccp._active = True

        ccp._fric = np.sqrt(ccp._bodya.fric * ccp._bodyb.fric)

        vcp: VelocityConstraintPoint = ccp._vcp
        vcp._ra = pair._pa - collision._bodya.pos
        vcp._rb = pair._pb - collision._bodyb.pos

        vcp._normal = collision._normal
        vcp._tangent = vcp._normal.perpendicular()

        im_a: float = collision._bodya.inv_mass
        im_b: float = collision._bodyb.inv_mass
        ii_a: float = collision._bodya.inv_inertia
        ii_b: float = collision._bodyb.inv_inertia

        rn_a: float = vcp._ra.cross(vcp._normal)
        rn_b: float = vcp._rb.cross(vcp._normal)

        rt_a: float = vcp._ra.cross(vcp._tangent)
        rt_b: float = vcp._rb.cross(vcp._tangent)

        k_normal: float = im_a + ii_a * rn_a * rn_a
        k_normal += im_b + ii_b * rn_b * rn_b
        k_tangent: float = im_a + ii_a * rt_a * rt_a
        k_tangent += im_b + ii_b * rt_b * rt_b

        vcp._eff_mass_normal = 0.0 if np.isclose(k_normal,
                                                 0) else 1.0 / k_normal
        vcp._eff_mass_tangent = 0.0 if np.isclose(k_tangent,
                                                  0) else 1.0 / k_tangent

        vcp._restit = np.fmin(ccp._bodya.restit, ccp._bodyb.restit)
        vcp._penetration = collision._penetration

        wa: Matrix = Matrix.cross_product2(ccp._bodya.ang_vel, vcp._ra)
        wb: Matrix = Matrix.cross_product2(ccp._bodyb.ang_vel, vcp._rb)
        vcp._va = ccp._bodya.vel + wa
        vcp._vb = ccp._bodyb.vel + wb

        vcp._vel_bias = (vcp._va - vcp._vb) * -vcp._restit

        # accumulate inherited impulse
        impulse: Matrix = vcp._normal * vcp._accum_normal_impulse
        impulse += vcp._tangent * vcp._accum_tangent_impulse

        ccp._bodya.apply_impulse(impulse, vcp._ra)
        ccp._bodyb.apply_impulse(-impulse, vcp._rb)

    def clear_inactive_points(self) -> None:
        clear_list: List[int] = []
        removed_list: List[ContactConstraintPoint] = []

        for key, val in self._contact_table.items():
            if len(val) == 0:
                clear_list.append(key)
                continue

            for v1 in val:
                if not v1._active:
                    removed_list.append(v1)

            for v1 in removed_list:
                for re in val:
                    if re == v1:
                        val.remove(re)

            removed_list.clear()

        for v2 in clear_list:
            for item in self._contact_table.items():
                if item[0] == v2:
                    del self._contact_table[item[0]]
                    break

    def deactivate_all_points(self) -> None:
        for val in self._contact_table.values():
            if len(val) == 0 or not val[0]._active:
                continue

            for ccp in val:
                ccp._active = False



import sys

import numpy as np
import taichi as ti

# add the TaichiGAME lib to the path
# below code is only needed in dev condition
sys.path.append('../')

import TaichiGAME as ng

ti.init(arch=ti.cpu)

scene = ng.Scene('CPU Testbed')


class FrameBroadPhaseDetect(ng.Frame):
    def load(self):
        tri_data = [
            ng.Matrix([-1.0, 1.0], 'vec'),
            ng.Matrix([0.0, -2.0], 'vec'),
            ng.Matrix([1.0, -1.0], 'vec'),
            ng.Matrix([-1.0, 1.0], 'vec'),
        ]

        poly_data = [
            ng.Matrix([0.0, 4.0], 'vec'),
            ng.Matrix([-3.0, 3.0], 'vec'),
            ng.Matrix([-4.0, 0.0], 'vec'),
            ng.Matrix([-3.0, -3.0], 'vec'),
            ng.Matrix([0.0, -4.0], 'vec'),
            ng.Matrix([3.0, -3.0], 'vec'),
            ng.Matrix([4.0, 0.0], 'vec'),
            ng.Matrix([3.0, 3.0], 'vec'),
            ng.Matrix([0.0, 4.0], 'vec')
        ]

        rect = ng.Rectangle(0.5, 0.5)
        cir = ng.Circle(0.5)
        cap = ng.Capsule(1.5, 0.5)
        tri = ng.Polygon()
        tri.vertices = tri_data
        poly = ng.Polygon()
        poly.vertices = poly_data

        tri.scale(0.5)
        poly.scale(0.1)

        rng = np.random.default_rng(seed=6)

        bd = ng.Body()
        tmpx: float = 0.0
        tmpy: float = 0.0
        cnt = np.array([])
        for i in range(51):
            bd = scene._world.create_body()
            tmpx = -10 + rng.random() * 20.0
            tmpy = -6 + rng.random() * 12.0
            bd.pos = ng.Matrix([tmpx, tmpy], 'vec')

            cnt = rng.integers(low=0, high=5, size=1)
            if cnt[0] == 0:
                bd.shape = rect
            elif cnt[0] == 1:
                bd.shape = cir
            elif cnt[0] == 2:
                bd.shape = tri
            elif cnt[0] == 3:
                bd.shape = poly
            elif cnt[0] == 4:
                bd.shape = cap

            bd.rot = -np.pi + rng.random() * np.pi
            bd.mass = 1
            bd.type = ng.Body.Type.Static
            scene._dbvt.insert(bd)

    def render(self):
        # render the query rectangle range
        ng.Render.rd_rect(
            scene._gui,
            scene._cam.world_to_screen(ng.Matrix([-4.0, 4.0], 'vec')),
            scene._cam.world_to_screen(ng.Matrix([4.0, -4.0], 'vec')),
            color=ng.Config.QueryRectLineColor)

        query_region = ng.AABB(8, 8)
        query_body_list = scene._dbvt.query(query_region)

        for bd in query_body_list:
            scene._cam.render_aabb(scene._gui, ng.AABB.from_body(bd))


class FrameRaycast(ng.Frame):
    def load(self):
        tri_data = [
            ng.Matrix([-1.0, 1.0], 'vec'),
            ng.Matrix([0.0, -2.0], 'vec'),
            ng.Matrix([1.0, -1.0], 'vec'),
            ng.Matrix([-1.0, 1.0], 'vec')
        ]

        poly_data = [
            ng.Matrix([0.0, 4.0], 'vec'),
            ng.Matrix([-3.0, 3.0], 'vec'),
            ng.Matrix([-4.0, 0.0], 'vec'),
            ng.Matrix([-3.0, -3.0], 'vec'),
            ng.Matrix([0.0, -4.0], 'vec'),
            ng.Matrix([3.0, -3.0], 'vec'),
            ng.Matrix([4.0, 0.0], 'vec'),
            ng.Matrix([3.0, 3.0], 'vec'),
            ng.Matrix([0.0, 4.0], 'vec')
        ]

        rect = ng.Rectangle(0.5, 0.5)
        cir = ng.Circle(0.5)
        cap = ng.Capsule(1.5, 0.5)
        tri = ng.Polygon()
        tri.vertices = tri_data
        poly = ng.Polygon()
        poly.vertices = poly_data

        tri.scale(0.5)
        poly.scale(0.1)

        rng = np.random.default_rng(seed=6)

        bd = ng.Body()
        tmpx: float = 0.0
        tmpy: float = 0.0
        cnt = np.array([])
        for i in range(51):
            bd = scene._world.create_body()
            tmpx = -10 + rng.random() * 20.0
            tmpy = -6 + rng.random() * 12.0
            bd.pos = ng.Matrix([tmpx, tmpy], 'vec')

            cnt = rng.integers(low=0, high=5, size=1)
            if cnt[0] == 0:
                bd.shape = rect
            elif cnt[0] == 1:
                bd.shape = cir
            elif cnt[0] == 2:
                bd.shape = tri
            elif cnt[0] == 3:
                bd.shape = poly
            elif cnt[0] == 4:
                bd.shape = cap

            bd.rot = -np.pi + rng.random() * np.pi
            bd.mass = 1
            bd.type = ng.Body.Type.Static
            scene._dbvt.insert(bd)

    def render(self):
        st = ng.Matrix([0.0, 0.0], 'vec')
        dirn = scene._mouse_pos.normal()

        ng.Render.rd_line(scene._gui,
                          scene._cam.world_to_screen(st),
                          scene._cam.world_to_screen(dirn * 10.0),
                          color=0xFF0000)
        query_body_list = scene._dbvt.raycast(st, dirn)

        prim = ng.ShapePrimitive()
        for bd in query_body_list:
            prim._rot = bd.rot
            prim._xform = bd.pos
            prim._shape = bd.shape
            ng.Render.rd_shape(scene._gui, prim, scene._cam.world_to_screen,
                               scene._cam.meter_to_pixel,
                               ng.Config.QueryRaycasFillColor,
                               ng.Config.QueryRaycasOutLineColor)


class FrameBitmask(ng.Frame):
    def load(self):
        rect = ng.Rectangle(1.0, 1.0)
        edg = ng.Edge()
        edg.set_value(ng.Matrix([-10.0, 0.0], 'vec'),
                      ng.Matrix([10.0, 0.0], 'vec'))

        mask: int = 0x01
        for i in range(3):
            grd = scene._world.create_body()
            grd.shape = edg
            grd.pos = ng.Matrix([0.0, -3.0 + i * 3.0], 'vec')
            grd.fric = 0.9
            grd.bitmask = mask
            grd.restit = 0
            grd.mass = ng.Config.Max
            grd.type = ng.Body.Type.Static
            mask <<= 1
            scene._dbvt.insert(grd)

        mask = 0x01
        for i in range(3):
            bd = scene._world.create_body()
            bd.shape = rect
            bd.pos = ng.Matrix([i * 3.0, 6.0], 'vec')
            bd.fric = 0.9
            bd.bitmask = mask
            bd.restit = 0.0
            bd.mass = 1
            bd.type = ng.Body.Type.Dynamic
            mask <<= 1
            scene._dbvt.insert(bd)

    def render(self):
        pass


class FrameCollision(ng.Frame):
    def load(self):
        edg = ng.Edge()
        edg.set_value(ng.Matrix([-10.0, 0.0], 'vec'),
                      ng.Matrix([10.0, 0.0], 'vec'))
        cap = ng.Capsule(2.0, 1.0)

        grd = scene._world.create_body()
        grd.shape = edg
        grd.pos = ng.Matrix([0.0, 0.0], 'vec')
        grd.mass = ng.Config.Max
        grd.type = ng.Body.Type.Static
        grd.fric = 0.7
        grd.restit = 1.0
        scene._dbvt.insert(grd)

        bd = scene._world.create_body()
        bd.shape = cap
        bd.pos = ng.Matrix([0.0, 6.0], 'vec')
        bd.rot = np.pi / 4
        bd.mass = 1
        bd.type = ng.Body.Type.Dynamic
        bd.fric = 0.4
        bd.restit = 0.0
        scene._dbvt.insert(bd)

    def render(self):
        pass


# BUG: some restit will lead to some enormous bias
class FrameRestitution(ng.Frame):
    def load(self):
        cir = ng.Circle(1.0)
        edg = ng.Edge()
        edg.set_value(ng.Matrix([-100.0, 0.0], 'vec'),
                      ng.Matrix([100.0, 0.0], 'vec'))

        grd = scene._world.create_body()
        grd.shape = edg
        grd.mass = ng.Config.Max
        grd.restit = 1.0
        grd.fric = 0.9
        grd.pos = ng.Matrix([0.0, 0.0], 'vec')
        grd.type = ng.Body.Type.Static
        scene._dbvt.insert(grd)

        for i in range(10):
            if i == 4 or i == 9:
                continue
            bd = scene._world.create_body()
            bd.shape = cir
            bd.mass = 10
            bd.fric = 0.1
            bd.restit = i / 10.0
            bd.pos = ng.Matrix([i * 2.5 - 10.0, 10.0], 'vec')
            bd.type = ng.Body.Type.Dynamic
            scene._dbvt.insert(bd)

    def render(self):
        pass


class FrameFriction(ng.Frame):
    def load(self):
        edg = ng.Edge()
        edg.set_value(ng.Matrix([0.0, 0.0], 'vec'),
                      ng.Matrix([10.0, 0.0], 'vec'))
        ramp = ng.Edge()
        ramp.set_value(ng.Matrix([-10.0, 4.0], 'vec'),
                       ng.Matrix([0.0, 0.0], 'vec'))
        rect = ng.Rectangle(0.5, 0.5)

        for i in range(3):
            grd = scene._world.create_body()
            grd.shape = edg
            grd.fric = 0.1
            grd.mass = ng.Config.Max
            grd.pos = ng.Matrix([0, i * 3.0], 'vec')
            grd.restit = 0.0
            grd.type = ng.Body.Type.Static

            ram_grd = scene._world.create_body()
            ram_grd.shape = ramp
            ram_grd.fric = 0.1
            ram_grd.mass = ng.Config.Max
            ram_grd.pos = ng.Matrix([0, i * 3.0], 'vec')
            ram_grd.restit = 0.0
            ram_grd.type = ng.Body.Type.Static

            scene._dbvt.insert(grd)
            scene._dbvt.insert(ram_grd)

        for i in range(1, 4):
            cube = scene._world.create_body()
            cube.shape = rect
            cube.fric = i * 0.3
            cube.mass = 1
            cube.pos = ng.Matrix([-5.0, i * 3.5], 'vec')
            cube.restit = 0.0
            cube.type = ng.Body.Type.Dynamic
            scene._dbvt.insert(cube)

    def render(self):
        pass


class FrameStack(ng.Frame):
    def load(self):
        edg = ng.Edge()
        edg.set_value(ng.Matrix([-100.0, 0.0], 'vec'),
                      ng.Matrix([100.0, 0.0], 'vec'))
        rect = ng.Rectangle(1.0, 1.0)
        grd = scene._world.create_body()
        grd.shape = edg
        grd.pos = ng.Matrix([0.0, 0.0], 'vec')
        grd.mass = ng.Config.Max
        grd.type = ng.Body.Type.Static
        scene._dbvt.insert(grd)

        offset: float = 0.0
        max: int = 4
        for i in range(max):
            for j in range(max - i):
                bd = scene._world.create_body()
                bd.pos = ng.Matrix([-10.0 + j * 1.1 + offset, i * 1.8 + 2.0],
                                   'vec')
                bd.shape = rect
                bd.rot = 0.0
                bd.mass = 0.2
                bd.type = ng.Body.Type.Dynamic
                bd.fric = 0.8
                bd.restit = 0.0
                scene._dbvt.insert(bd)

            offset += 0.5

    def render(self):
        pass


# BUG: some impl error
class FrameBridge(ng.Frame):
    def load(self):
        brick = ng.Rectangle(1.5, 0.5)
        edg = ng.Edge()
        edg.set_value(ng.Matrix([-100.0, 0.0], 'vec'),
                      ng.Matrix([100.0, 0], 'vec'))

        rect = scene._world.create_body()
        rect.shape = brick
        rect.pos = ng.Matrix([-15.0, 0.0], 'vec')
        rect.rot = 0.0
        rect.mass = 1.0
        rect.restit = 0.2
        rect.fric = 0.01
        rect.type = ng.Body.Type.Dynamic

        grd = scene._world.create_body()
        grd.shape = edg
        grd.pos = ng.Matrix([0.0, -15.0], 'vec')
        grd.mass = ng.Config.Max
        grd.type = ng.Body.Type.Static
        scene._dbvt.insert(grd)

        ppm = ng.PointJointPrimitive()
        revol = ng.RevoluteJointPrimitive()

        half: float = brick.width / 2.0
        ppm._bodya = rect
        ppm._local_pointa.set_value([-half, 0.0])
        ppm._target_point.set_value([-15.0 - half, 0.0])
        ppm._damping_radio = 0.1
        ppm._freq = 1000
        ppm._force_max = 10000
        scene._world.create_joint(ppm)

        cnt: int = 20
        scene._dbvt.insert(rect)

        for i in range(1, cnt):
            rect2 = scene._world.create_body()
            rect2.shape = brick
            rect2.pos = ng.Matrix([-15.0 + i * brick.width * 1.2, 0.0], 'vec')
            rect2.rot = 0.0
            rect2.mass = 1.0
            rect2.fric = 0.01
            rect2.type = ng.Body.Type.Dynamic

            scene._dbvt.insert(rect2)
            revol._bodya = rect
            revol._bodyb = rect2
            revol._local_pointa.set_value([half + brick.width * 0.1, 0.0])
            revol._local_pointb.set_value([-half - brick.width * 0.1, 0.0])
            revol._damping_radio = 0.8
            revol._freq = 10.0
            revol._force_max = 10000
            scene._world.create_joint(revol)
            rect = rect2

        ppm._bodya = rect2
        ppm._local_pointa.set_value([0.75, 0.0])
        ppm._target_point.set_value(
            rect2.to_world_point(ng.Matrix([0.75, 0.0], 'vec')))
        ppm._damping_radio = 0.1
        ppm._freq = 1000.0
        ppm._force_max = 10000.0
        scene._world.create_joint(ppm)

    def render(self):
        pass


class FrameDomino(ng.Frame):
    def load(self):
        floor = ng.Rectangle(15.0, 0.5)
        rect = ng.Rectangle(0.5, 0.5)
        brick = ng.Rectangle(0.3, 3.0)
        edg = ng.Edge()
        edg.set_value(ng.Matrix([-100.0, 0.0], 'vec'),
                      ng.Matrix([100.0, 0.0], 'vec'))

        grd = scene._world.create_body()
        grd.shape = edg
        grd.type = ng.Body.Type.Static
        grd.mass = ng.Config.Max
        grd.pos = ng.Matrix([0.0, 0.0], 'vec')
        grd.fric = 0.1
        grd.restit = 0.0
        scene._dbvt.insert(grd)

        tile = scene._world.create_body()
        tile.shape = floor
        tile.type = ng.Body.Type.Static
        tile.mass = ng.Config.Max
        tile.fric = 0.1
        tile.restit = 0.0
        tile.rot = np.pi / 180 * 15
        tile.pos = ng.Matrix([4.0, 8.0], 'vec')
        scene._dbvt.insert(tile)

        tile = scene._world.create_body()
        tile.shape = floor
        tile.type = ng.Body.Type.Static
        tile.mass = ng.Config.Max
        tile.fric = 0.1
        tile.restit = 0.0
        tile.rot = np.pi / 180 * -15
        tile.pos = ng.Matrix([-4.0, 4.0], 'vec')
        scene._dbvt.insert(tile)

        tile = scene._world.create_body()
        tile.shape = floor
        tile.type = ng.Body.Type.Static
        tile.mass = ng.Config.Max
        tile.fric = 0.1
        tile.restit = 0.0
        tile.rot = 0.0
        tile.pos = ng.Matrix([-5.0, 10.0], 'vec')
        scene._dbvt.insert(tile)

        for i in range(9):
            card = scene._world.create_body()
            card.shape = brick
            card.mass = 10
            card.fric = 0.1
            card.restit = 0.0
            card.type = ng.Body.Type.Dynamic
            card.pos = ng.Matrix([-10.0 + i * 1.2, 12.0], 'vec')
            scene._dbvt.insert(card)

        pendulum = scene._world.create_body()
        pendulum.shape = rect
        pendulum.mass = 2.0
        pendulum.fric = 0.1
        pendulum.type = ng.Body.Type.Dynamic
        pendulum.pos = ng.Matrix([-16.0, 16.0], 'vec')
        scene._dbvt.insert(pendulum)

        djp = ng.DistanceJointPrimitive()
        djp._bodya = pendulum
        djp._local_pointa.set_value([0.0, 0.0])
        djp._dist_min = 1.0
        djp._dist_max = 4.0
        djp._target_point.set_value([-12.0, 16.0])
        scene._world.create_joint(djp)

        # ojp: OrientationJointPrimitive = OrientationJointPrimitive()
        # ojp._target_point.set_value([-12.0, 16.0])
        # ojp._bodya = pendulum
        # ojp._ref_rot = 0.0
        # scene._world.create_joint(ojp)

    def render(self):
        pass


frame_broad_phase = FrameBroadPhaseDetect()
frame_raycast = FrameRaycast()
frame_bitmask = FrameBitmask()
frame_collision = FrameCollision()
frame_restitution = FrameRestitution()
frame_fricitoin = FrameFriction()
frame_stack = FrameStack()
frame_bridge = FrameBridge()
frame_domino = FrameDomino()

scene.register_frame(frame_broad_phase)
scene.register_frame(frame_raycast)
scene.register_frame(frame_bitmask)
scene.register_frame(frame_collision)
scene.register_frame(frame_restitution)
scene.register_frame(frame_fricitoin)
scene.register_frame(frame_stack)
scene.register_frame(frame_domino)
scene.init_frame()

scene.show()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="TaichiGAME",
    version="0.0.3",
    author="maksyuki",
    author_email="maksyuki@126.com",
    description="GPU Accelerated Motion Engine based on Taichi Lang",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/maksyuki/TaichiGAME",
    project_urls={
        'Documentation': 'https://github.com/maksyuki/TaichiGAME',
        'Funding': 'https://donate.pypi.org',
        'Say Thanks!': 'http://saythanks.io/to/example',
        'Source': 'https://github.com/maksyuki/TaichiGAME',
        'Tracker': 'https://github.com/maksyuki/TaichiGAME/issues',
    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License', 'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Framework :: Robot Framework :: Library',
        'Topic :: Games/Entertainment :: Simulation',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Software Development :: Libraries :: Application Frameworks'
    ],
    license='MIT',
    keywords=['phyics engine', 'dynamics simulation', 'robot motion control'],
    packages=setuptools.find_packages(exclude=['tests']),
    include_package_data=True,
    install_requires=['taichi'],
    python_requires=">=3.7,<3.10",
)



from __future__ import annotations
from typing import Dict, Tuple, Union, List, cast, Optional

import taichi as ti

from .common.camera import Camera
from .common.config import Config
from .common.export_manager import ExportManager
from .frame import Frame
from .collision.broad_phase.dbvt import DBVT
from .collision.broad_phase.aabb import AABB
from .math.matrix import Matrix
from .collision.detector import Collsion, Detector
from .dynamics.body import Body
from .dynamics.phy_world import PhysicsWorld
from .dynamics.constraint.contact import ContactMaintainer
from .dynamics.joint.point import PointJoint, PointJointPrimitive


class Scene():
    def __init__(self,
                 name: str,
                 width: int = 1280,
                 height: int = 720,
                 option: Dict[str, bool] = {}):
        self._gui: ti.GUI = ti.GUI(name,
                                   res=(width, height),
                                   background_color=Config.BackgroundColor)
        self._option: Dict[str, bool] = {'video': False, 'gif': False}
        for v in option:
            self._option[v] = option[v]
        self._ex_mgn: ExportManager = ExportManager()
        # the physics world, all sim is run in physics world
        self._world: PhysicsWorld = PhysicsWorld()
        self._dbvt: DBVT = DBVT()
        self._maintainer: ContactMaintainer = ContactMaintainer()
        # the view camera, all viewport scale is in camera
        self._cam: Camera = Camera()

        # physics init settings
        self._world.grav = Matrix([0.0, -9.8], 'vec')
        self._world.damping_ena = True
        self._world._linear_vel_damping = 0.1
        self._world.ang_vel_damping = 0.1
        self._world.air_fric_coeff = 0.8
        self._world.pos_iter = 8
        self._world.vel_iter = 6

        # camera init settings
        self._cam.viewport = Camera.Viewport(Matrix([0.0, height], 'vec'),
                                             Matrix([width, 0.0], 'vec'))
        self._cam.body_visible = True
        self._cam.center_visible = True
        self._cam.rot_line_visible = True
        self._cam.joint_visible = True
        # self._cam.aabb_visible = True
        # self._cam.dbvt_visible = True
        self._cam._world = self._world
        self._cam._dbvt = self._dbvt

        # extern frame table
        self._ext_frame_list: List[Frame] = []
        self._ext_frame_idx: int = 0

        # calcuate step
        self._fps = 120
        self._dt = 1 / self._fps
        self._paused = False
        # NOTE: some algorithm need to cacluate the pos's len
        # in init state
        self._mouse_pos: Matrix = Matrix([1.0, 1.0], 'vec')
        # the right-mouse btn drag move flag(change viewport)
        self._mouse_viewport_move: bool = False

        # mouse joint oper
        self._mouse_joint_prim: PointJointPrimitive = PointJointPrimitive()
        self._mouse_joint_prim._bodya = Body()
        self._mouse_joint: PointJoint = cast(
            PointJoint, self._world.create_joint(self._mouse_joint_prim))
        self._mouse_joint.active = False
        self._mouse_select_body: Optional[Body] = None

    def register_frame(self, frame: Frame) -> None:
        self._ext_frame_list.append(frame)

    def remove_frame(self, frame: Frame) -> None:
        # NOTE: need to check if exist first
        self._ext_frame_list.remove(frame)

    def clear_all(self) -> None:
        self._world.clear_all_bodies()
        self._world.clear_all_joints()
        self._maintainer.clear_all()
        self._dbvt.clear_all()
        self._mouse_joint_prim._bodya = Body()
        self._mouse_joint = cast(
            PointJoint, self._world.create_joint(self._mouse_joint_prim))
        self._mouse_joint.active = False

    def calc_nxt_frame(self, delta: int) -> None:
        ext_len: int = len(self._ext_frame_list)
        assert -ext_len <= delta <= ext_len

        self._ext_frame_idx = Config.clamp(self._ext_frame_idx, 0, ext_len - 1)
        # NOTE: the sign of mod in python depend on dividend
        self._ext_frame_idx = (self._ext_frame_idx + delta) % ext_len

    def init_frame(self) -> None:
        self.change_frame(0)

    def change_frame(self, delta: int) -> None:
        self.clear_all()
        self.calc_nxt_frame(delta)
        self._ext_frame_list[self._ext_frame_idx].load()

    def physics_sim(self) -> None:
        for elem in self._world._body_list:
            self._dbvt.update(elem)

        self._world.step_velocity(self._dt)

        pot_list: List[Tuple[Body, Body]] = self._dbvt.generate()
        # print('sim')
        for pot in pot_list:
            # print('pot')
            # print(f'bodya: ({pot[0].pos.x}, {pot[0].pos.y}), {pot[0].rot}')
            # print(f'bodyb: ({pot[1].pos.x}, {pot[1].pos.y}), {pot[1].rot}')

            res: Collsion = Detector.detect(pot[0], pot[1])
            if res._is_colliding:
                # print('collid')
                # print(f'contact list len: {len(res._contact_list)}')
                self._maintainer.add(res)

        self._maintainer.clear_inactive_points()
        self._world.prepare_velocity_constraint(self._dt)

        for i in range(self._world.vel_iter):
            self._world.solve_velocity_constraint(self._dt)
            self._maintainer.solve_velocity(self._dt)

        self._world.step_position(self._dt)

        for i in range(self._world.pos_iter):
            self._maintainer.solve_position(self._dt)
            self._world.solve_position_constraint(self._dt)

        self._maintainer.deactivate_all_points()

    def render(self) -> None:
        self._cam.render(self._gui)
        self._ext_frame_list[self._ext_frame_idx].render()

    def handle_left_mouse_event(self, state: Union[ti.GUI.PRESS,
                                                   ti.GUI.RELEASE], x: float,
                                y: float) -> None:

        self._mouse_pos = self._cam.screen_to_world(Matrix([x, y], 'vec'))
        if state == ti.GUI.PRESS:
            if self._mouse_joint is None:
                return

            mouse_box: AABB = AABB(0.01, 0.01)
            mouse_box.pos = self._mouse_pos
            bd_list: List[Body] = self._dbvt.query(mouse_box)
            for bd in bd_list:
                # print(bd.id)
                point: Matrix = self._mouse_pos - bd.pos
                point = Matrix.rotate_mat(-bd.rot) * point

                if bd.shape.contains(
                        point) and self._mouse_select_body is None:
                    self._mouse_select_body = bd
                    prim: PointJointPrimitive = self._mouse_joint.prim()
                    prim._local_pointa = bd.to_local_point(self._mouse_pos)
                    prim._bodya = bd
                    prim._target_point = self._mouse_pos
                    self._mouse_joint.active = True
                    self._mouse_joint.set_value(prim)
                    break
        else:
            self._mouse_joint.active = False
            self._mouse_select_body = None

    def handle_right_mouse_event(
            self, state: Union[ti.GUI.PRESS, ti.GUI.RELEASE]) -> None:
        if state == ti.GUI.PRESS:
            self._mouse_viewport_move = True
        else:
            self._mouse_viewport_move = False

    def handle_mouse_move_event(self, x: float, y: float) -> None:
        cur_pos: Matrix = self._cam.screen_to_world(Matrix([x, y], 'vec'))
        delta_pos: Matrix = cur_pos - self._mouse_pos

        if self._mouse_viewport_move:
            # print(f'delta_pos1: {delta_pos}')
            delta_pos *= self._cam.meter_to_pixel
            # print(f'delta_pos2: {delta_pos}')
            # 0.5 just a hack value for equal radio move offset
            # 33.0 is init meter_to_pixel value
            self._cam.transform += delta_pos * 0.5 * (33.0 /
                                                      self._cam.meter_to_pixel)

        self._mouse_pos = cur_pos
        # print(f'({self._mouse_pos.x}, {self._mouse_pos.y}')
        if self._mouse_joint is None:
            return

        prim: PointJointPrimitive = self._mouse_joint.prim()
        prim._target_point = self._mouse_pos
        self._mouse_joint.set_value(prim)

    def handle_wheel_event(self, y: float) -> None:
        # NOTE: need to set the sef._cam  the value scale
        # zoom in
        if y > 0:
            self._cam.meter_to_pixel += self._cam.meter_to_pixel / 4
        else:
            self._cam.meter_to_pixel -= self._cam.meter_to_pixel / 4

    def show(self) -> None:
        while self._gui.running:

            for e in self._gui.get_events():
                if e.key == ti.GUI.ESCAPE:
                    if self._option['gif']:
                        self._ex_mgn.gen_gif()
                    elif self._option['video']:
                        self._ex_mgn.gen_video()

                    exit()

                elif e.key == ti.GUI.SPACE and e.type == ti.GUI.RELEASE:
                    self._paused = not self._paused

                elif e.key == ti.GUI.LMB:
                    self.handle_left_mouse_event(e.type, e.pos[0], e.pos[1])

                elif e.key == ti.GUI.RMB:
                    self.handle_right_mouse_event(e.type)

                elif e.key == ti.GUI.MOVE:
                    self.handle_mouse_move_event(e.pos[0], e.pos[1])

                elif e.key == ti.GUI.WHEEL:
                    self.handle_wheel_event(e.delta[1])

                elif e.key == ti.GUI.UP:
                    print("press up key")

                elif e.key == ti.GUI.DOWN:
                    pass

                elif e.key == ti.GUI.LEFT and e.type == ti.GUI.RELEASE:
                    self.change_frame(-1)

                elif e.key == ti.GUI.RIGHT and e.type == ti.GUI.RELEASE:
                    self.change_frame(1)

                elif e.key == 'q' and e.type == ti.GUI.PRESS:
                    self._cam.visible = not self._cam.visible

                elif e.key == 'w' and e.type == ti.GUI.PRESS:
                    self._cam.aabb_visible = not self._cam.aabb_visible

                elif e.key == 'e' and e.type == ti.GUI.PRESS:
                    self._cam.joint_visible = not self._cam.joint_visible

                elif e.key == 'r' and e.type == ti.GUI.PRESS:
                    self._cam.body_visible = not self._cam.body_visible

                elif e.key == 't' and e.type == ti.GUI.PRESS:
                    self._cam.axis_visible = not self._cam.axis_visible

                elif e.key == 'a' and e.type == ti.GUI.PRESS:
                    self._cam.dbvh_visible = not self._cam.dbvh_visible

                elif e.key == 's' and e.type == ti.GUI.PRESS:
                    self._cam.dbvt_visible = not self._cam.dbvt_visible

                elif e.key == 'd' and e.type == ti.GUI.PRESS:
                    self._cam.grid_visible = not self._cam.grid_visible

                elif e.key == 'f' and e.type == ti.GUI.PRESS:
                    self._cam.rot_line_visible = not self._cam.rot_line_visible

                elif e.key == 'g' and e.type == ti.GUI.PRESS:
                    self._cam.center_visible = not self._cam.center_visible

                elif e.key == 'z' and e.type == ti.GUI.PRESS:
                    self._cam.contact_visible = not self._cam.contact_visible

            if not self._paused:
                self.physics_sim()

            self.render()

            if self._option['video'] or self._option['gif']:
                self._gui.show(self._gui.show(self._ex_mgn.frame_name))
            else:
                self._gui.show()
